import io
import logging
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor


import boto3
import botocore

logger = logging.getLogger(__name__)


class FileSystemManager:
    """Methods for interacting with files."""

    def __init__(self, config):
        self.key_prefix = (
            None  # sertting a default value to fix undefined variable error
        )
        self.config = config

        self.aws_endpoint_url = self.config.get("aws_endpoint_url")
        self.replacement_host = None

        self.client_config = botocore.config.Config(
            max_pool_connections=100,
            signature_version=self.config.get("LQS_SIG_VER", "s3v4"),
        )

        self.store_session = True
        self.session = None
        self.clients = {}
        self.resources = {}

    # S3 Methods

    def set_store_session(self, value):
        self.store_session = value

    def get_boto_session(self):
        if self.session is not None:
            return self.session
        session = boto3.session.Session(
            aws_access_key_id=self.config.get("aws_access_key_id"),
            aws_secret_access_key=self.config.get("aws_secret_access_key"),
            region_name=self.config.get("aws_region"),
            aws_session_token=self.config.get("aws_session_token"),
        )
        if self.store_session:
            self.session = session
        return session

    def get_boto_client(self, service_name, session=None, replacement_host=None):
        if replacement_host is None:
            if service_name in self.clients:
                return self.clients[service_name]
        else:
            if f"{service_name}:{replacement_host}" in self.clients:
                return self.clients[f"{service_name}:{replacement_host}"]

        if session is None:
            session = self.get_boto_session()
        endpoint_url = self.aws_endpoint_url
        if replacement_host is not None:
            endpoint_url = replacement_host
        client = session.client(
            service_name=service_name,
            endpoint_url=endpoint_url,
            config=self.client_config,
        )

        if self.store_session:
            if replacement_host is None:
                self.clients[service_name] = client
            else:
                self.clients[f"{service_name}:{replacement_host}"] = client
        return client

    def get_boto_resource(self, service_name, session=None):
        if service_name in self.resources:
            return self.resources[service_name]
        if session is None:
            session = self.get_boto_session()
        resource = session.resource(
            service_name=service_name,
            endpoint_url=self.aws_endpoint_url,
            config=self.client_config,
        )
        if self.store_session:
            self.resources[service_name] = resource
        return resource

    def prepare_s3_key(self, s3_key):
        if not s3_key.startswith(self.key_prefix or ""):
            s3_key = self.key_prefix + s3_key
        return s3_key

    def get_data(self, s3_bucket, s3_key, start_offset=None, end_offset=None):
        s3_resource = self.get_boto_resource("s3")
        s3_object = s3_resource.Object(bucket_name=s3_bucket, key=s3_key)
        if start_offset is not None and end_offset is not None:
            range_header = f"bytes={start_offset}-{end_offset}"
            return s3_object.get(Range=range_header)["Body"].read()
        else:
            return s3_object.get()["Body"].read()

    def download_file(self, s3_bucket, s3_key, filepath):
        s3_client = self.get_boto_client("s3")
        download_dir = "/".join(filepath.split("/")[:-1])
        Path(download_dir).mkdir(parents=True, exist_ok=True)
        response = s3_client.download_file(s3_bucket, s3_key, filepath)
        logger.debug(response)
        return response

    def put_data(self, data, s3_bucket, s3_key, tags=None):
        s3_client = self.get_boto_client("s3")
        params = dict(Body=data, Bucket=s3_bucket, Key=s3_key)
        if tags is not None and self.enable_tagging:
            tagging = "&".join([f"{k}={v}" for k, v in tags.items()])
            params["Tagging"] = tagging
        response = s3_client.put_object(**params)
        logger.debug(response)
        return response

    def upload_file(self, filepath, s3_bucket, s3_key):
        s3_client = self.get_boto_client("s3")
        response = s3_client.upload_file(filepath, s3_bucket, s3_key)
        logger.debug(response)
        return response

    def upload_fileobj(self, data, s3_bucket, s3_key):
        s3_client = self.get_boto_client("s3")
        response = s3_client.upload_fileobj(data, s3_bucket, s3_key)
        logger.debug(response)
        return response

    def head_data(self, s3_bucket, s3_key, fail_on_404=False):
        try:
            s3_client = self.get_boto_client("s3")
            response = s3_client.head_object(Bucket=s3_bucket, Key=s3_key)
            logger.debug(response)
            return response
        except botocore.exceptions.ClientError as e:
            if not fail_on_404 and e.response["Error"]["Code"] == "404":
                return None
            else:
                raise e

    def list_data(self, s3_bucket, s3_prefix):
        s3_client = self.get_boto_client("s3")
        response = s3_client.list_objects_v2(Bucket=s3_bucket, Prefix=s3_prefix)
        logger.debug(response)
        return response

    def delete_data(self, s3_bucket, s3_key):
        s3_client = self.get_boto_client("s3")
        response = s3_client.delete_object(Bucket=s3_bucket, Key=s3_key)
        logger.debug(response)
        return response

    def get_object_size(self, s3_bucket, s3_key):
        s3_resource = self.get_boto_resource("s3")
        s3_object = s3_resource.Object(bucket_name=s3_bucket, key=s3_key)
        return s3_object.content_length

    def create_bucket(self, bucket_name):
        s3_resource = self.get_boto_resource("s3")
        return s3_resource.create_bucket(Bucket=bucket_name)

    def generate_presigned_url(self, params, client_method="get_object"):
        s3_client = self.get_boto_client("s3", replacement_host=self.replacement_host)
        url = s3_client.generate_presigned_url(
            ClientMethod=client_method, Params=params, ExpiresIn=self.url_expiration
        )
        if self.replacement_host is not None:
            url = url.replace(self.aws_endpoint_url, self.replacement_host)
        logger.debug(f"presigned url: {url[:10]}...{url[-5:]}")
        return url

    def get_object_chunk(
        self, barray, s3_object, range_header, data_offset, chunk_size
    ):
        response = s3_object.get(Range=range_header)
        barray[data_offset : data_offset + chunk_size] = response["Body"].read()

    def get_object_body(self, s3_bucket, s3_key, offset=0, size=-1):
        s3_resource = self.get_boto_resource("s3")
        s3_object = s3_resource.Object(bucket_name=s3_bucket, key=s3_key)

        offset = int(offset)
        size = int(size)
        if size < 2e9:
            # if it's small enough, just download it directly
            if size > 0:
                data = bytearray(size)
            else:
                data = bytearray(s3_object.content_length - offset)
            range_header = f"bytes={offset}-"
            if size > 0:
                range_end = str(offset + int(size - 1))
                range_header += range_end
            response = s3_object.get(Range=range_header)
            data[:] = response["Body"].read()
        else:
            data = bytearray(size)
            data_offset = 0
            download_size = int(1.28e8)  # 128 MB
            max_workers = 4
            # download_size = int(1e9) # 1 GB
            chunk_count = int(size // download_size)
            chunk_offset = offset
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                for chunk_idx in range(chunk_count + 1):
                    chunk_size = int(
                        min(download_size, size - (chunk_idx * download_size))
                    )
                    range_header = (
                        f"bytes={chunk_offset}-"
                        f"{str(chunk_offset + int(chunk_size - 1))}"
                    )
                    if int(chunk_size - 1) < 1:
                        logger.warn(f"Skipping last byte: {range_header}")
                        continue
                    logger.info(
                        (
                            f"Downloading Chunk {chunk_idx}/{chunk_count}: "
                            f"{range_header}"
                        )
                    )
                    chunk_offset += chunk_size
                    executor.submit(
                        self.get_object_chunk,
                        data,
                        s3_object,
                        range_header,
                        data_offset,
                        chunk_size,
                    )
                    data_offset += chunk_size
        return data

    def create_presigned_url(self, method, params):
        return {
            "method": method,
            "params": params,
            "url": self.generate_presigned_url(params=params, client_method=method),
        }



import logging

import fire

from . import LogQS

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s  (%(levelname)s - %(name)s) : %(message)s"
)
logger = logging.getLogger(__name__)


if __name__ == "__main__":
    fire.Fire(LogQS)



import struct
import io

# The magic bits to indicate the byte is a header.
BITSTREAM_HEADER_MAGIC = 0xD0

# THe magic bits to indicate the byte is a group.
BITSTREAM_GROUP_MAGIC = 0xE0

# Indicates that this header/group has at least one following section
# that needs to be read.
BITSTREAM_HAS_FOLLOWING_SECTION = 0x04


class InputStream:
    """
    This class allows you to read serialized rbuf structures into native
    Python structures. It closely mimics the C++ API.
    """

    #
    # This block of structures is a pre-initialized/static set of structures meant
    # to improve performance of decoding individual primitives.
    #

    _decoder_uint8 = struct.Struct("<B")
    _decoder_uint16 = struct.Struct("<H")
    _decoder_uint32 = struct.Struct("<L")
    _decoder_uint64 = struct.Struct("<Q")
    _decoder_int8 = struct.Struct("<b")
    _decoder_int16 = struct.Struct("<h")
    _decoder_int32 = struct.Struct("<l")
    _decoder_int64 = struct.Struct("<q")
    _decoder_float = struct.Struct("<f")
    _decoder_double = struct.Struct("<d")
    _decoder_group = struct.Struct("<BBL")

    def __init__(self, buffer):
        """
        Initializes the stream against the given buffer of bytes.
        """

        self.buffer = io.BytesIO(buffer)
        self.has_more_sections = False

    def read_bitstream_header(self):
        """
        Reads the bitstream header from the stream. These headers
        denote the start of a new serialized structure, and help
        indicate version information or if different sections are
        present.
        """
        # print(f"/read_bitstream_header: {self.buffer.tell()}")
        header = self.read_uint8()

        # Ensure the header magic bits are set.
        if (header & 0xF0) != BITSTREAM_HEADER_MAGIC:
            raise Exception(
                "Tried to read a bitstream header, but the header had the incorrect magic numbers. "
                "This indicates the stream is corrupt, or is not a rbuf stream."
            )

        # We can only decode version one with this code.
        if (header & 0x03) != 0x01:
            raise Exception(
                "Tried to read a bitstream header, but it had an unexpected version number. This "
                "software only supports version 1."
            )

        # Store off if we have more sections...
        self.has_more_sections = (header & BITSTREAM_HAS_FOLLOWING_SECTION) > 0

        # If the last bit is set, then we have some kind of unexpected
        # bitstream error.
        if (header & 0x08) != 0x00:
            raise Exception(
                "Tried to read a bitstream header, but this header indicates the stream has groups, "
                "which this software does not support."
            )

    def read_optional_group_header(self):
        """
        Reads a group header from the bitstream, updating if we have additional sections or not,
        and returning a tuple containing the parsed identifier and group size. Throws if the next
        byte does not appear to be a group header.
        """
        # print(f"read_optional_group_header: {self.buffer.tell()}")
        header = InputStream._decoder_group.unpack(self.buffer.read(6))

        # Throw if this isn't a group
        if (header[0] & 0xF0) != BITSTREAM_GROUP_MAGIC:
            raise Exception(
                "Tried to read a bitstream group header, but the header had an incorrect magic "
                "number. This indicates the stream is corrupt, or is not an rbuf stream."
            )

        # Record if we have additional sections
        self.has_more_sections = (header[0] & BITSTREAM_HAS_FOLLOWING_SECTION) > 0

        # Throw if we have any other fields set, something is wrong and we're an incorrect
        # version or were wrong that it was a group.
        if (header[0] & 0x0B) != 0x00:
            raise Exception(
                "Tried to read a bitstream group header, but the header had additional flags specified that were unknown."
            )

        # Return the group identifier and size as a tuple
        return header[1:]

    def advance(self, length):
        """
        Advances the bitstream by length bytes, throwing if this would go outside of the bounds
        of the buffer. This can be used to 'read' data from the stream (skip it).
        """

        self.buffer.seek(length, 1)

    def read_string(self):
        """
        Reads the next string from the input stream, throwing if there is
        a bitstream error.
        """

        # Read the string length first, it's a 32-bit integer.
        string_length = self.read_uint32()

        # The next string_length bytes are returned as the actual string data...
        string = self.buffer.read(string_length)

        return string.decode("utf-8")

    def read_object(self, object_class):
        """
        Reads the next object from the input stream. The object class is used to instantiate
        the object.
        """

        # Push state before deserializing
        pushed_flags = self.has_more_sections
        self.has_more_sections = False

        # Create the object and deserialize
        output = object_class()
        output.deserialize(self)

        # Restore state and return the final result
        self.has_more_sections = pushed_flags
        return output

    def read_variant(self, types):
        """
        Reads a variant from the given list of possibilities.
        """

        index = self.read_uint8()
        length = self.read_uint32()

        # If the index isn't in our types, skip it, and return the first type
        # to match C++ logic
        if index not in types:
            self.advance(length)

            entries = list(types.values())
            return entries[0]()

        # Now we can read the object given the passed-in type.
        return self.read_object(types[index])

    def read_array_size(self, fixed_array_size):
        """
        Reads the array size and returns it. If fixed_array_size is not none, just
        returns that value.
        """

        if fixed_array_size is None:
            return self.read_uint32()
        else:
            return fixed_array_size

    def read_object_array(self, object_class, fixed_array_size=None):
        """
        Reads an array containing the given object class.

        If fixed_array_size is specified, assumes the array size is fixed, and uses that
        number instead of reading the length from the bitsteam.
        """

        # Get the desired array length...
        array_length = self.read_array_size(fixed_array_size)

        # Create the array and read objects into it.
        output = []

        while array_length > 0:
            output.append(self.read_object(object_class))
            array_length -= 1

        return output

    def read_array(self, read_fn, fixed_array_size=None):
        """
        Reads an array with the given function lambda.

        If fixed_array_size is specified, assumes the array size is fixed, and uses that
        number instead of reading the length from the bitsteam.
        """

        # Get the desired array length...
        array_length = self.read_array_size(fixed_array_size)

        # Create the array and read objects into it.
        output = []

        while array_length > 0:
            output.append(read_fn())
            array_length -= 1

        return output

    def read_primitive_array(self, format_string, item_length, fixed_array_size=None):
        """
        Reads a primitive array of the given format, checking to make sure we have
        enough bytes left in the bitstream to satisfy the request.

        If fixed_array_size is specified, assumes the array size is fixed, and uses that
        number instead of reading the length from the bitsteam.
        """

        # Read the length of the array and check to make sure we have enough room
        # for the content.
        array_length = self.read_array_size(fixed_array_size)
        byte_size = array_length * item_length

        # Now read the actual array
        read_data = struct.unpack(
            "<" + str(array_length) + format_string, self.buffer.read(byte_size)
        )

        return list(read_data)

    def read_byte_buffer(self):
        """
        Reads a 'byte buffer' (in Python, this is just an array of uint8 objects).
        """

        # Read the length in first...
        array_length = self.read_uint32()

        # Now just read the raw bytes. No need to do any packing/unpacking/etc, this is just
        # an arary of bytes.
        return self.buffer.read(array_length)

    def read_guid(self):
        """
        Reads a GUID from the bitstream, which is essentially a 16-byte array that we
        return into a string for Python consumption.
        """

        # Read the raw 16 bytes in.
        raw_bytes = self.buffer.read(16)

        # Convert the array into a hex guid, the hard way. Note that in Ark
        # the first 8 bytes are behind the second 8 bytes, and 'backwards', as
        # it tries to match little endian notation for two 8 byte integers for
        # some reason. We should move this into a class.
        output = ""

        index = 7
        while index >= 0:
            if index == 1 or index == 3:
                output += "-"

            output += "{:02x}".format(raw_bytes[index])
            index -= 1

        index = 15
        while index >= 8:
            if index == 15 or index == 13:
                output += "-"

            output += "{:02x}".format(raw_bytes[index])
            index -= 1

        return output

    def read_dictionary(self, key_type_fn, value_type_fn, dictionary_length=None):
        """
        Reads in a dictioanry. The key and value function actually read data from
        the underlying bitstream and place the contents into said dictionary.
        """

        dictionary_length = self.read_array_size(dictionary_length)
        result = {}
        # print(f"dictionary_length: {dictionary_length}")

        while dictionary_length > 0:
            key = key_type_fn()
            value = value_type_fn()

            result[key] = value
            dictionary_length -= 1

        return result

    def read_optional(self, read_fn):
        """
        Reads in an optional. If set, uses the read function to read the primitive,
        otherwise, returns None.
        """

        if self.read_bool():
            return read_fn()
        else:
            return None

    def read_bool(self):
        return self.read_uint8() > 0

    def read_uint8(self):
        return InputStream._decoder_uint8.unpack(self.buffer.read(1))[0]

    def read_uint16(self):
        return InputStream._decoder_uint16.unpack(self.buffer.read(2))[0]

    def read_uint32(self):
        return InputStream._decoder_uint32.unpack(self.buffer.read(4))[0]

    def read_uint64(self):
        return InputStream._decoder_uint64.unpack(self.buffer.read(8))[0]

    def read_int8(self):
        return InputStream._decoder_int8.unpack(self.buffer.read(1))[0]

    def read_int16(self):
        return InputStream._decoder_int16.unpack(self.buffer.read(2))[0]

    def read_int32(self):
        return InputStream._decoder_int32.unpack(self.buffer.read(4))[0]

    def read_int64(self):
        return InputStream._decoder_int64.unpack(self.buffer.read(8))[0]

    def read_float(self):
        return InputStream._decoder_float.unpack(self.buffer.read(4))[0]

    def read_double(self):
        return InputStream._decoder_double.unpack(self.buffer.read(8))[0]



import os
import logging
from distutils.util import strtobool

from dotenv import dotenv_values

logging.basicConfig(
    level=os.getenv("LQS_LOG_LEVEL") or logging.INFO,
    format="%(asctime)s  (%(levelname)s - %(name)s): %(message)s",
)
logger = logging.getLogger(__name__)

from .interface import (
    Lister,
    Getter,
    Creator,
    Updater,
    Deleter,
    S3,
    Utils,
    Node,
    Terminal,
)
from .gen import MessageGenerator


class LogQSClient:
    def __init__(
        self,
        api_url=None,
        api_key_id=None,
        api_key_secret=None,
        pretty=None,
        verbose=None,
        dry_run=None,
        retry_count=None,
        retry_delay=None,
        retry_aggressive=None,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        aws_region=None,
        aws_session_token=None,
        aws_endpoint_url=None,
        verify_ssl=True,
    ):
        self._env_config = {**dotenv_values(".env"), **os.environ}

        self._api_url = api_url if api_url else self._env_config.get("LQS_API_URL")
        if not self._api_url:
            raise ValueError("No API URL provided")
        self._api_key_id = (
            api_key_id if api_key_id else self._env_config.get("LQS_API_KEY_ID")
        )
        if not self._api_key_id:
            raise ValueError("No API Key ID provided")
        self._api_key_secret = (
            api_key_secret
            if api_key_secret
            else self._env_config.get("LQS_API_KEY_SECRET")
        )
        if not self._api_key_secret:
            raise ValueError("No API Key Secret provided")

        self._pretty = (
            pretty
            if pretty
            else bool(strtobool(self._env_config.get("LQS_PRETTY", "false")))
        )

        self._verbose = (
            verbose
            if verbose
            else bool(strtobool(self._env_config.get("LQS_VERBOSE", "false")))
        )

        if self._verbose:
            logger.setLevel(logging.DEBUG)

        self._dry_run = (
            dry_run
            if dry_run
            else bool(strtobool(self._env_config.get("LQS_DRY_RUN", "false")))
        )

        self._retry_count = (
            retry_count
            if retry_count is not None
            else int(self._env_config.get("LQS_RETRY_COUNT", 4))
        )
        if self._retry_count < 0:
            raise ValueError("Retry count must be greater than or equal to 0")

        self._retry_delay = (
            retry_delay
            if retry_delay is not None
            else int(self._env_config.get("LQS_RETRY_DELAY", 4))
        )
        if self._retry_delay < 0:
            raise ValueError("Retry delay must be greater than or equal to 0")

        self._retry_aggressive = (
            retry_aggressive
            if retry_aggressive is not None
            else bool(strtobool(self._env_config.get("LQS_RETRY_AGGRESSIVE", "false")))
        )

        self._aws_access_key_id = (
            aws_access_key_id
            if aws_access_key_id is not None
            else self._env_config.get("AWS_ACCESS_KEY_ID")
        )
        self._aws_secret_access_key = (
            aws_secret_access_key
            if aws_secret_access_key is not None
            else self._env_config.get("AWS_SECRET_ACCESS_KEY")
        )
        self._aws_region = (
            aws_region
            if aws_region is not None
            else self._env_config.get("AWS_DEFAULT_REGION")
        )
        self._aws_session_token = (
            aws_session_token
            if aws_session_token is not None
            else self._env_config.get("AWS_SESSION_TOKEN")
        )
        self._aws_endpoint_url = (
            aws_endpoint_url
            if aws_endpoint_url is not None
            else self._env_config.get("AWS_ENDPOINT_URL")
        )

        self._verify_ssl = verify_ssl

        self.config = {
            "api_url": self._api_url,
            "api_key_id": self._api_key_id,
            "api_key_secret": self._api_key_secret,
            "pretty": self._pretty,
            "verbose": self._verbose,
            "dry_run": self._dry_run,
            "retry_count": self._retry_count,
            "retry_delay": self._retry_delay,
            "retry_aggressive": self._retry_aggressive,
            "aws_access_key_id": self._aws_access_key_id,
            "aws_secret_access_key": self._aws_secret_access_key,
            "aws_region": self._aws_region,
            "aws_session_token": self._aws_session_token,
            "aws_endpoint_url": self._aws_endpoint_url,
            "verify_ssl": self._verify_ssl,
        }

        safe_config = {k: v for k, v in self.config.items()}

        obf_key = self._api_key_secret[:4] + "*" * (len(self._api_key_secret) - 4)
        safe_config["api_key_secret"] = obf_key

        if self._aws_secret_access_key:
            obf_aws_key = self._aws_secret_access_key[:4] + "*" * (
                len(self._aws_secret_access_key) - 4
            )
            safe_config["aws_secret_access_key"] = obf_aws_key

        logger.debug("config: %s", safe_config)

        self.list = Lister(config=self.config)
        self.get = Getter(config=self.config)
        self.create = Creator(config=self.config)
        self.update = Updater(config=self.config)
        self.delete = Deleter(config=self.config)

        self.s3 = S3(config=self.config, getter=self.get, creator=self.create)
        self.gen = MessageGenerator(getter=self.get, lister=self.list)
        self.utils = Utils(
            getter=self.get, s3=self.s3, gen=self.gen, config=self.config
        )
        self.node = Node(
            getter=self.get, lister=self.list, utils=self.utils, config=self.config
        )
        self.terminal = Terminal(getter=self.get, lister=self.list, config=self.config)


LogQS = LogQSClient



from typing import Optional
from datetime import datetime
from uuid import UUID

from .common import RESTInterface, ProcessStatus, UserRole, LogFormat, output_decorator


class Lister(RESTInterface):
    def __init__(self, config):
        super().__init__(config)

    @output_decorator
    def api_key(
        self,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "created_at",
        sort: Optional[str] = "ASC",
        user_id: Optional[UUID] = None,
        key: Optional[str] = None,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"apiKeys" + self._get_url_param_string(args, [])
        result = self._get_resource(resource_path)
        return result

    @output_decorator
    def extraction(
        self,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "created_at",
        sort: Optional[str] = "ASC",
        log_id: Optional[UUID] = None,
        group_id: Optional[UUID] = None,
        status: Optional[ProcessStatus] = None,
        processing: Optional[bool] = None,
        queued: Optional[bool] = None,
        cancelled: Optional[bool] = None,
        errored: Optional[bool] = None,
        archived: Optional[bool] = None,
        completed: Optional[bool] = None,
        name: Optional[str] = None,
        name_like: Optional[str] = None,
        format: Optional[LogFormat] = None,
        s3_bucket: Optional[str] = None,
        s3_key: Optional[str] = None,
        s3_key_prefix: Optional[str] = None,
        size_null: Optional[bool] = None,
        size_gte: Optional[int] = None,
        size_lte: Optional[int] = None,
        progress_null: Optional[bool] = None,
        progress_gte: Optional[float] = None,
        progress_lte: Optional[float] = None,
        error_like: Optional[str] = None,
        note_like: Optional[str] = None,
        context_filter: Optional[dict] = None,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"extractions" + self._get_url_param_string(args, [])
        result = self._get_resource(resource_path)
        return result

    @output_decorator
    def extraction_topic(
        self,
        extraction_id: UUID,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "created_at",
        sort: Optional[str] = "ASC",
        topic_id: Optional[UUID] = None,
        start_time_null: Optional[bool] = None,
        start_time_gte: Optional[float] = None,
        start_time_lte: Optional[float] = None,
        end_time_null: Optional[bool] = None,
        end_time_gte: Optional[float] = None,
        end_time_lte: Optional[float] = None,
        size_null: Optional[bool] = None,
        size_gte: Optional[int] = None,
        size_lte: Optional[int] = None,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"extractions/{extraction_id}" + self._get_url_param_string(
            args, ["extraction_id"]
        )
        result = self._get_resource(resource_path)
        return result

    @output_decorator
    def group(
        self,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "created_at",
        sort: Optional[str] = "ASC",
        name: Optional[str] = None,
        name_like: Optional[str] = None,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"groups" + self._get_url_param_string(args, [])
        result = self._get_resource(resource_path)
        return result

    @output_decorator
    def group_association(
        self,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "created_at",
        sort: Optional[str] = "ASC",
        user_id: Optional[UUID] = None,
        group_id: Optional[UUID] = None,
        role: Optional[UserRole] = None,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"groupAssociations" + self._get_url_param_string(args, [])
        result = self._get_resource(resource_path)
        return result

    @output_decorator
    def ingestion(
        self,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "created_at",
        sort: Optional[str] = "ASC",
        log_id: Optional[UUID] = None,
        group_id: Optional[UUID] = None,
        status: Optional[ProcessStatus] = None,
        processing: Optional[bool] = None,
        queued: Optional[bool] = None,
        cancelled: Optional[bool] = None,
        errored: Optional[bool] = None,
        archived: Optional[bool] = None,
        completed: Optional[bool] = None,
        format: Optional[LogFormat] = None,
        name: Optional[str] = None,
        name_like: Optional[str] = None,
        s3_bucket: Optional[str] = None,
        s3_key: Optional[str] = None,
        s3_key_prefix: Optional[str] = None,
        size_null: Optional[bool] = None,
        size_gte: Optional[int] = None,
        size_lte: Optional[int] = None,
        progress_null: Optional[bool] = None,
        progress_gte: Optional[float] = None,
        progress_lte: Optional[float] = None,
        meta_like: Optional[str] = None,
        error_like: Optional[str] = None,
        note_like: Optional[str] = None,
        context_filter: Optional[dict] = None,
        start_offset_null: Optional[bool] = None,
        start_offset_gte: Optional[int] = None,
        start_offset_lte: Optional[int] = None,
        end_offset_null: Optional[bool] = None,
        end_offset_gte: Optional[int] = None,
        end_offset_lte: Optional[int] = None,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"ingestions" + self._get_url_param_string(args, [])
        result = self._get_resource(resource_path)
        return result

    @output_decorator
    def log(
        self,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "created_at",
        sort: Optional[str] = "ASC",
        group_id: Optional[UUID] = None,
        locked: Optional[bool] = None,
        start_time_null: Optional[bool] = None,
        start_time_gte: Optional[float] = None,
        start_time_lte: Optional[float] = None,
        end_time_null: Optional[bool] = None,
        end_time_gte: Optional[float] = None,
        end_time_lte: Optional[float] = None,
        name: Optional[str] = None,
        name_like: Optional[str] = None,
        note_like: Optional[str] = None,
        context_filter: Optional[dict] = None,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"logs" + self._get_url_param_string(args, [])
        result = self._get_resource(resource_path)
        return result

    @output_decorator
    def message_type(
        self,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "created_at",
        sort: Optional[str] = "ASC",
        group_id: Optional[UUID] = None,
        name: Optional[str] = None,
        name_like: Optional[str] = None,
        md5: Optional[str] = None,
        definition_like: Optional[str] = None,
        note_like: Optional[str] = None,
        context_filter: Optional[dict] = None,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"messageTypes" + self._get_url_param_string(args, [])
        result = self._get_resource(resource_path)
        return result

    @output_decorator
    def record(
        self,
        topic_id: UUID,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "timestamp",
        sort: Optional[str] = "ASC",
        log_id: Optional[UUID] = None,
        ingestion_id: Optional[UUID] = None,
        errored: Optional[bool] = None,
        archived: Optional[bool] = None,
        data_filter: Optional[dict] = None,
        context_filter: Optional[dict] = None,
        frequency: Optional[float] = None,
        timestamp_gt: Optional[float] = None,
        timestamp_lt: Optional[float] = None,
        timestamp_gte: Optional[float] = None,
        timestamp_lte: Optional[float] = None,
        length_gte: Optional[int] = None,
        length_lte: Optional[int] = None,
        data_length_gte: Optional[int] = None,
        data_length_lte: Optional[int] = None,
        offset_gte: Optional[int] = None,
        offset_lte: Optional[int] = None,
        data_offset_gte: Optional[int] = None,
        data_offset_lte: Optional[int] = None,
        nanosecond_gte: Optional[int] = None,
        nanosecond_lte: Optional[int] = None,
        error_like: Optional[str] = None,
        note_like: Optional[str] = None,
        include_image: Optional[bool] = False,
        include_bytes: Optional[bool] = False,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"topics/{topic_id}/records" + self._get_url_param_string(
            args, []
        )
        result = self._get_resource(resource_path)
        return result

    @output_decorator
    def topic(
        self,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "created_at",
        sort: Optional[str] = "ASC",
        log_id: Optional[UUID] = None,
        group_id: Optional[UUID] = None,
        message_type_id: Optional[UUID] = None,
        associated_topic_id: Optional[UUID] = None,
        ingestion_id: Optional[UUID] = None,
        locked: Optional[bool] = None,
        name: Optional[str] = None,
        name_like: Optional[str] = None,
        context_filter: Optional[dict] = None,
        type_name: Optional[str] = None,
        type_name_like: Optional[str] = None,
        type_encoding: Optional[str] = None,
        type_encoding_like: Optional[str] = None,
        type_data: Optional[str] = None,
        type_data_like: Optional[str] = None,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"topics" + self._get_url_param_string(args, [])
        result = self._get_resource(resource_path)
        return result

    @output_decorator
    def user(
        self,
        offset: Optional[int] = 0,
        limit: Optional[int] = 10,
        order: Optional[str] = "created_at",
        sort: Optional[str] = "ASC",
        username: Optional[str] = None,
        username_like: Optional[str] = None,
        is_admin: Optional[bool] = None,
        created_by: Optional[UUID] = None,
        updated_by: Optional[UUID] = None,
        deleted_by: Optional[UUID] = None,
        updated_by_null: Optional[bool] = None,
        deleted_by_null: Optional[bool] = None,
        updated_at_null: Optional[bool] = None,
        deleted_at_null: Optional[bool] = None,
        created_at_lte: Optional[datetime] = None,
        updated_at_lte: Optional[datetime] = None,
        deleted_at_lte: Optional[datetime] = None,
        created_at_gte: Optional[datetime] = None,
        updated_at_gte: Optional[datetime] = None,
        deleted_at_gte: Optional[datetime] = None,
    ):
        args = locals()
        resource_path = f"users" + self._get_url_param_string(args, [])
        result = self._get_resource(resource_path)
        return result

    # Plural methods
    def api_keys(self, **kwargs):
        return self.api_key(**kwargs)

    def extractions(self, **kwargs):
        return self.extraction(**kwargs)

    def extraction_topics(self, **kwargs):
        return self.extraction_topic(**kwargs)

    def groups(self, **kwargs):
        return self.group(**kwargs)

    def group_associations(self, **kwargs):
        return self.group_association(**kwargs)

    def ingestions(self, **kwargs):
        return self.ingestion(**kwargs)

    def logs(self, **kwargs):
        return self.log(**kwargs)

    def message_types(self, **kwargs):
        return self.message_type(**kwargs)

    def records(self, **kwargs):
        return self.record(**kwargs)

    def topics(self, **kwargs):
        return self.topic(**kwargs)

    def users(self, **kwargs):
        return self.user(**kwargs)



import os
import struct
import time
import logging
from uuid import UUID
from enum import Enum
from typing import Optional, Union, List

logging.basicConfig(
    level=os.getenv("LQS_LOG_LEVEL") or logging.INFO,
    format="%(asctime)s  (%(levelname)s - %(name)s): %(message)s",
)
logger = logging.getLogger(__name__)

import lz4.frame
import requests
import xmltodict
from urllib.parse import unquote

import lzma
import zstd
import lz4.block
import bz2
import struct
import s3path

from .fs import FileSystemManager


class S3Method(str, Enum):
    abort_multipart_upload = "abort_multipart_upload"
    complete_multipart_upload = "complete_multipart_upload"
    create_multipart_upload = "create_multipart_upload"
    delete_object = "delete_object"
    get_object = "get_object"
    head_object = "head_object"
    list_multipart_uploads = "list_multipart_uploads"
    list_object_versions = "list_object_versions"
    list_objects_v2 = "list_objects_v2"
    list_parts = "list_parts"
    put_object = "put_object"
    restore_object = "restore_object"
    upload_part = "upload_part"


class S3Resource(str, Enum):
    extraction = "extraction"
    ingestion = "ingestion"
    log = "log"
    record = "record"
    topic = "topic"


class S3:
    def __init__(self, config, getter, creator):
        self._config = config
        self._getter = getter
        self._creator = creator

        self._retry_count = config["retry_count"]
        self._retry_delay = config["retry_delay"]
        self._retry_aggressive = config["retry_aggressive"]
        self._verify_ssl = config["verify_ssl"]

        self.fs = FileSystemManager(config)

    def _handle_retries(self, func, retry_count=None):
        if retry_count is None:
            retry_count = self._retry_count
        for i in range(retry_count + 1):
            try:
                return func()
            except Exception as e:
                if not self._retry_aggressive:
                    lqs_expected_error_codes = [
                        "[BadRequest]",
                        "[Forbidden]",
                        "[NotFound]",
                        "[Conflict]",
                        "[Locked]",
                    ]
                    for code in lqs_expected_error_codes:
                        if code in str(e):
                            raise e
                if retry_count > 0 and i < retry_count:
                    # exponential backoff
                    backoff = self._retry_delay * (2**i)
                    logger.error(f"Error: {e}")
                    logger.info(f"Retrying in {backoff} seconds")
                    time.sleep(backoff)
                else:
                    raise e

    def generate_presigned_url(
        self,
        method: S3Method,
        resource: S3Resource,
        params: dict,
        resource_id: UUID = None,
        timestamp: Optional[float] = None,
    ):
        if resource == "extraction":
            presigned_url = self._creator.extraction_presigned_url(
                extraction_id=resource_id, method=method, params=params
            )
        elif resource == "ingestion":
            presigned_url = self._creator.ingestion_presigned_url(
                ingestion_id=resource_id, method=method, params=params
            )
        elif resource == "log":
            presigned_url = self._creator.log_presigned_url(
                log_id=resource_id, method=method, params=params
            )
        elif resource == "record":
            if timestamp is None:
                raise ValueError("timestamp is required.")
            presigned_url = self._creator.record_presigned_url(
                topic_id=resource_id, timestamp=timestamp, method=method, params=params
            )
        elif resource == "topic":
            presigned_url = self._creator.topic_presigned_url(
                topic_id=resource_id, method=method, params=params
            )
        else:
            raise ValueError(f"Unknown resource {resource}.")
        url = presigned_url["url"]
        used_params = presigned_url["params"]
        return url, used_params

    def abort_multi_part_upload(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        key: str,
        upload_id: str,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {
            "Key": key,
            "UploadId": upload_id,
        }

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="abort_multipart_upload",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )

            r = requests.delete(url, verify=self._verify_ssl)
            return r.headers, used_params, None

        return self._handle_retries(make_requests)

    def complete_multipart_upload(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        key: str,
        upload_id: str,
        parts: List[dict],
        convert_to_dict: bool = True,
        decode_response: bool = True,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {
            "Key": key,
            "UploadId": upload_id,
        }

        complete_multipart_payload = "<CompleteMultipartUpload>"
        for part in parts:
            if "ETag" not in part or "PartNumber" not in part:
                raise ValueError("Part must contain 'ETag' and 'PartNumber'.")
            part["ETag"] = part["ETag"].strip('"')
            complete_multipart_payload += f"""
            <Part>
                <PartNumber>{part["PartNumber"]}</PartNumber>
                <ETag>{part["ETag"]}</ETag>
            </Part>
            """
        complete_multipart_payload += "</CompleteMultipartUpload>"

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="complete_multipart_upload",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.post(
                url, data=complete_multipart_payload, verify=self._verify_ssl
            )
            if r.status_code != 200:
                raise ValueError(f"Error completing multipart upload: {r.text}")
            if decode_response:
                response_text = unquote(r.text)
            else:
                response_text = r.text
            response_content = (
                xmltodict.parse(response_text) if convert_to_dict else response_text
            )
            return r.headers, used_params, response_content

        return self._handle_retries(make_requests)

    def create_multipart_upload(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        key: str,
        convert_to_dict: bool = True,
        decode_response: bool = True,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {"Key": key}

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="create_multipart_upload",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.post(url, verify=self._verify_ssl)
            if r.status_code != 200:
                raise ValueError(f"Error creating multipart upload: {r.text}")
            if decode_response:
                response_text = unquote(r.text)
            else:
                response_text = r.text
            response_content = (
                xmltodict.parse(response_text) if convert_to_dict else response_text
            )
            return r.headers, used_params, response_content

        return self._handle_retries(make_requests)

    def delete_object(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        key: str,
        version_id: Optional[str] = None,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {"Key": key}
        if version_id is not None:
            params["VersionId"] = version_id

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="delete_object",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.delete(url, verify=self._verify_ssl)
            if r.status_code != 204:
                raise ValueError(f"Error deleting object: {r.text}")
            return r.headers, used_params, None

        return self._handle_retries(make_requests)

    def get_object(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        key: str,
        part_number: Optional[int] = None,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {"Key": key}
        if part_number is not None:
            params["PartNumber"] = part_number

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="get_object",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.get(url, verify=self._verify_ssl)
            if r.status_code != 200:
                raise ValueError(f"Error getting object: {r.text}")
            return r.headers, used_params, r.content

        return self._handle_retries(make_requests)

    def head_object(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        key: str,
        part_number: Optional[int] = None,
        version_id: Optional[str] = None,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {"Key": key}
        if part_number is not None:
            params["PartNumber"] = part_number
        if version_id is not None:
            params["VersionId"] = version_id

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="head_object",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.head(url, verify=self._verify_ssl)
            # TODO: may need to return a status code as well
            # for now, we'll just return it as the body
            return r.headers, used_params, {"exists": r.status_code == 200}

        return self._handle_retries(make_requests)

    def list_multipart_uploads(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        convert_to_dict: bool = True,
        decode_response: bool = True,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {}

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="list_multipart_uploads",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.get(url, verify=self._verify_ssl)
            if r.status_code != 200:
                raise ValueError(f"Error listing multipart uploads: {r.text}")
            if decode_response:
                response_text = unquote(r.text)
            else:
                response_text = r.text
            response_content = (
                xmltodict.parse(response_text) if convert_to_dict else response_text
            )
            return r.headers, used_params, response_content

        return self._handle_retries(make_requests)

    def list_object_versions(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        convert_to_dict: bool = True,
        decode_response: bool = True,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {}

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="list_object_versions",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.get(url, verify=self._verify_ssl)
            if r.status_code != 200:
                raise ValueError(f"Error listing object versions: {r.text}")
            if decode_response:
                response_text = unquote(r.text)
            else:
                response_text = r.text
            response_content = (
                xmltodict.parse(response_text) if convert_to_dict else response_text
            )
            return r.headers, used_params, response_content

        return self._handle_retries(make_requests)

    def list_objects_v2(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        continuation_token: Optional[str] = None,
        convert_to_dict: bool = True,
        decode_response: bool = True,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {}
        if continuation_token is not None:
            params["ContinuationToken"] = continuation_token

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="list_objects_v2",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.get(url, verify=self._verify_ssl)
            if r.status_code != 200:
                raise ValueError(f"Error listing objects: {r.text}")
            if decode_response:
                response_text = unquote(r.text)
            else:
                response_text = r.text
            response_content = (
                xmltodict.parse(response_text) if convert_to_dict else response_text
            )
            return r.headers, used_params, response_content

        return self._handle_retries(make_requests)

    def list_parts(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        key: str,
        upload_id: str,
        part_number_marker: Optional[int] = None,
        convert_to_dict: bool = True,
        decode_response: bool = True,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {"Key": key, "UploadId": upload_id}
        if part_number_marker is not None:
            params["PartNumberMarker"] = part_number_marker

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="list_parts",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.get(url, verify=self._verify_ssl)
            if r.status_code != 200:
                raise ValueError(f"Error listing parts: {r.text}")
            if decode_response:
                response_text = unquote(r.text)
            else:
                response_text = r.text
            response_content = (
                xmltodict.parse(response_text) if convert_to_dict else response_text
            )
            return r.headers, used_params, response_content

        return self._handle_retries(make_requests)

    def put_object(
        self, resource: S3Resource, resource_id: Union[str, dict], key: str, body: bytes
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {"Key": key}

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="put_object",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.put(url, data=body, verify=self._verify_ssl)
            if r.status_code != 200:
                raise ValueError(f"Error putting object: {r.text}")
            return r.headers, used_params, None

        return self._handle_retries(make_requests)

    def restore_object(self):
        raise NotImplementedError("Restore object is not implemented yet.")

    def upload_part(
        self,
        resource: S3Resource,
        resource_id: Union[str, dict],
        key: str,
        body: bytes,
        part_number: int,
        upload_id: str,
    ):
        if resource in ["record"]:
            if (
                isinstance(resource_id, str)
                or "topic_id" not in resource_id
                or "timestamp" not in resource_id
            ):
                raise ValueError(
                    "When resource is 'record', resource_id must be a dict with values 'topic_id' and 'timestamp'."
                )
            timestamp = resource_id["timestamp"]
            resource_id = resource_id["topic_id"]
        else:
            timestamp = None
        params = {"Key": key, "PartNumber": part_number, "UploadId": upload_id}

        def make_requests():
            url, used_params = self.generate_presigned_url(
                method="upload_part",
                resource=resource,
                params=params,
                resource_id=resource_id,
                timestamp=timestamp,
            )
            r = requests.put(url, data=body, verify=self._verify_ssl)
            if r.status_code != 200:
                raise ValueError(f"Error uploading part: {r.text}")
            return r.headers, used_params, None

        return self._handle_retries(make_requests)

    def get_message_data_from_record(
        self, record, s3_bucket=None, s3_key=None, ingestion_id=None
    ):
        if record.get("bytes_url") is not None:
            # we've requested the bytes url, so we want to use that
            return requests.get(record["bytes_url"], verify=self._verify_ssl).content

        if s3_bucket is None or s3_key is None:
            s3_bucket = record["s3_bucket"]
            s3_key = record["s3_key"]
            if s3_bucket is None or s3_key is None:
                ingestion_id = record["ingestion_id"]
                if ingestion_id is None:
                    raise ValueError("Missing ingestion_id and s3_bucket/s3_key")
                ingestion = self._getter.ingestion(ingestion_id=ingestion_id)["data"]
                s3_bucket = ingestion["s3_bucket"]
                s3_key = ingestion["s3_key"]

        chunk_compression = record["chunk_compression"]
        data_offset = record["data_offset"]
        data_length = record["data_length"]
        if chunk_compression != "none" and chunk_compression is not None:
            chunk_offset = record["chunk_offset"]
            chunk_length = record["chunk_length"]

            chunk_data = self.fs.get_data(
                s3_bucket=s3_bucket,
                s3_key=s3_key,
                start_offset=chunk_offset,
                end_offset=chunk_offset + chunk_length,
            )

            if chunk_compression == "lz4":
                decompressed_message_data = lz4.frame.decompress(chunk_data)
            else:
                raise ValueError(f"Unsupported chunk_compression: {chunk_compression}")

            message_data = decompressed_message_data[
                data_offset : data_offset + data_length
            ]
        else:
            message_data = self.fs.get_data(
                s3_bucket=s3_bucket,
                s3_key=s3_key,
                start_offset=data_offset,
                end_offset=data_offset + data_length,
            )
        return message_data

    def get_message_data_for_ark_record(
        self, record, s3_bucket=None, s3_key=None, ingestion_id=None
    ):
        if record.get("bytes_url") is not None:
            # we've requested the bytes url, so we want to use that
            return requests.get(record["bytes_url"], verify=self._verify_ssl).content

        if s3_bucket is None or s3_key is None:
            # TODO: abstract this logic (redundant in get_message_data_from_record)
            s3_bucket = record["s3_bucket"]
            s3_key = record["s3_key"]

            if s3_bucket is None or s3_key is None:
                ingestion_id = record["ingestion_id"]
                if ingestion_id is None:
                    raise ValueError("Missing ingestion_id and s3_bucket/s3_key")
                ingestion = self._getter.ingestion(ingestion_id=ingestion_id)["data"]
                s3_bucket = ingestion["s3_bucket"]
                if s3_key is None:
                    s3_key = ingestion["s3_key"]
        split_id = None
        if record["context"]:
            split_id = record["context"].get("split_id", None)

        if split_id:
            s3_key = self.get_split_s3_key(s3_key, split_id)
        else:
            assert record["s3_key"] is not None
            relative_s3_key = record["s3_key"]
            # get the manifest s3 key from the ingestion
            ingestion_id = record["ingestion_id"]
            ingestion = self._getter.ingestion(ingestion_id=ingestion_id)["data"]
            s3_key = ingestion["s3_key"]
            # combine the manifest s3 key with the relative s3 key
            s3_key = self._get_split_s3_key_for_record(
                manifest_s3_key=s3_key,
                relative_split_path=relative_s3_key,
                s3_bucket=s3_bucket,
            )

        compression = record["chunk_compression"]
        chunk_start_offset = record["chunk_offset"]
        # the offset is inclusive, so we need to subtract 1
        chunk_end_offset = record["chunk_offset"] + record["chunk_length"] - 1
        bytes = self.fs.get_data(
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            start_offset=chunk_start_offset,
            end_offset=chunk_end_offset,
        )
        message_bytes = self._decompress_ark_record(
            compression_type=compression, compressed_bytes=bytes
        )
        return message_bytes

    def _decompress_ark_record(
        self, compression_type: str, compressed_bytes: bytes
    ) -> bytes:
        if compression_type is None or compression_type.lower() == "none":
            decompressed_bytes = compressed_bytes
        else:
            compressed_footer = compressed_bytes[-6:]
            # size: uint32, compression_type: uint8, footer_version: uint8
            """
            None = 0,
            Lz4 = 1,
            Zstd = 2,
            Bz2 = 3,
            Lzma = 4
            """
            FOOTER_MAGIC = 0xF0
            (size, footer_compression_type, footer_version) = struct.unpack(
                "<LBB", compressed_footer
            )
            if footer_version != FOOTER_MAGIC:
                assert (
                    footer_version == FOOTER_MAGIC
                ), f"Footer version {footer_version} is not supported, only {FOOTER_MAGIC} is supported"

            elif compression_type == "lz4":
                assert (
                    footer_compression_type == 1
                ), f"Mismatch between compression type {compression_type} and footer compression type {footer_compression_type}"
                decompressed_bytes = lz4.block.decompress(compressed_bytes[:-6], size)
            elif compression_type == "zstd":
                assert (
                    footer_compression_type == 2
                ), f"Mismatch between compression type {compression_type} and footer compression type {footer_compression_type}"
                decompressed_bytes = zstd.decompress(compressed_bytes[:-6])
            elif compression_type == "bz2":
                assert (
                    footer_compression_type == 3
                ), f"Mismatch between compression type {compression_type} and footer compression type {footer_compression_type}"
                decompressed_bytes = bz2.decompress(compressed_bytes[:-6])
            elif compression_type == "lzma":
                assert (
                    footer_compression_type == 4
                ), f"Mismatch between compression type {compression_type} and footer compression type {footer_compression_type}"
                decompressed_bytes = lzma.decompress(compressed_bytes[:-6])
            else:
                raise Exception(f"Unsupported compression type {compression_type}")
        return decompressed_bytes

    def get_split_s3_key(self, manifest_s3_key: str, split_id: str) -> str:
        parent_path = manifest_s3_key.rsplit("/manifests", 1)[0]
        split_file_path = f"{parent_path}/splits/{split_id}"
        return self.fs.prepare_s3_key(split_file_path)

    def _get_split_s3_key_for_record(
        self, manifest_s3_key: str, relative_split_path: str, s3_bucket: str
    ) -> str:
        log_prefix = manifest_s3_key.rsplit("/manifests", 1)[0]
        split_s3_key = s3path.PureS3Path.from_bucket_key(
            bucket=s3_bucket, key=manifest_s3_key + "/" + relative_split_path
        ).key
        # if the split_s3_key doesn't start with the log_prefix, then something funky is going on
        if not split_s3_key.startswith(log_prefix):
            raise Exception(
                f"Split s3 key {split_s3_key} does not start with log prefix {log_prefix}"
            )

        return self.fs.prepare_s3_key(split_s3_key)



import os
import time
import logging
import threading

import requests
import rospy

logging.basicConfig(
    level=os.getenv("LQS_LOG_LEVEL") or logging.INFO,
    format="%(asctime)s  (%(levelname)s - %(name)s): %(message)s",
)
logger = logging.getLogger(__name__)

# http://wiki.ros.org/ROS/Tutorials/WritingPublisherSubscriber%28python%29


class Node:
    def __init__(self, getter, lister, utils, config):
        self._get = getter
        self._list = lister
        self._utils = utils
        self._config = config

        self._log = None
        self._start_time = None
        self._end_time = None

        self._publishers = {}
        self._subscriptions = {}
        self._remaps = {}

        if self._config.get("aws_access_key_id"):
            self._use_s3_directly = True
        else:
            self._use_s3_directly = False
        self._kill_threads = False

        if self._config.get("verbose"):
            logger.setLevel(logging.DEBUG)

    def subscribe(self, log_id, topic_names, rel_start_time, rel_end_time, remaps={}):
        rospy.init_node("lqs_client", anonymous=True)

        self._log = self._get.log(log_id=log_id)["data"]

        self._start_time = self._log["start_time"] + rel_start_time
        self._end_time = self._log["start_time"] + rel_end_time

        topics_all = self._list.topics(log_id=log_id, limit=100)["data"]
        self._remaps = remaps
        self._subscriptions = {}
        for topic in topics_all:
            if topic["name"] in topic_names:
                message_class = self._utils.get_message_class(topic["message_type_id"])
                self._publishers[topic["name"]] = rospy.Publisher(
                    remaps.get(topic["name"], topic["name"]),
                    message_class,
                    queue_size=10,
                )
                thread = threading.Thread(
                    target=self.load_records, args=(topic["name"],)
                )
                thread.start()
                self._subscriptions[topic["name"]] = {
                    "topic": topic,
                    "thread": thread,
                    "loading_data": True,  # assume there is data to start
                    "records": [],
                }

    def fetch_record_bytes(self, topic_name, record):
        record["fetching_bytes"] = True
        record["bytes"] = None
        start_time = time.time()
        if self._use_s3_directly:
            mode = "direct"
            record["bytes"] = self._utils.get_message_data_from_record(record)
        else:
            mode = "presgined"
            record["bytes"] = requests.get(record["bytes_url"]).content
        end_time = time.time()
        logger.debug(
            f"({mode}) fetched bytes for {topic_name} ({record['timestamp']}) in {end_time - start_time}"
        )

    def load_records(self, topic_name):
        limit = 100 if self._use_s3_directly else 10
        offset = 0
        while True:
            if topic_name not in self._subscriptions:
                continue
            sub = self._subscriptions[topic_name]
            records = sub["records"]
            topic = sub["topic"]
            records_response = self._list.records(
                log_id=self._log["id"],
                topic_id=topic["id"],
                timestamp_gte=self._start_time,
                timestamp_lt=self._end_time,
                limit=limit,
                offset=offset,
                include_bytes=False if self._use_s3_directly else True,
            )
            records = records_response["data"]
            logger.debug(f"Loaded {len(records)} records for {topic_name}")
            if len(records) == 0:
                break
            offset += limit
            sub["records"] += records
            if self._kill_threads:
                break
        self._subscriptions[topic_name]["loading_data"] = False
        logger.debug(f"Done loading records for {topic_name}.")

    def publish(self, topic_name):
        sub = self._subscriptions[topic_name]
        topic = sub["topic"]
        records = sub["records"]
        if len(records) == 0:
            logger.debug(f"No records for {topic_name}")
            return

        record = records.pop(0)
        if "fetching_bytes" not in record:
            self.fetch_record_bytes(topic_name, record)

        while record["bytes"] is None:
            time.sleep(0.01)

        if len(records) > 0:
            next_record = records[0]
            if "fetching_bytes" not in next_record:
                threading.Thread(
                    target=self.fetch_record_bytes, args=(topic_name, next_record)
                ).start()

        message = self._utils.get_ros_message_from_message_data(
            message_data=record["bytes"],
            message_type_id=topic["message_type_id"],
        )
        publisher = self._publishers[topic_name]
        publisher.publish(message)

    def publish_next_message(self):
        # look through our records and publish the next message based on timestamp
        next_message_time = None
        next_message_topic = None
        still_loading = False
        for topic_name in self._subscriptions:
            sub = self._subscriptions[topic_name]
            if sub["loading_data"]:
                still_loading = True
            if len(sub["records"]) == 0:
                if still_loading:
                    # no records, but we're still loading data, so we restart the loop
                    return True
                continue
            next_record = sub["records"][0]
            if (
                next_message_time is None
                or next_record["timestamp"] < next_message_time
            ):
                # this topic may have the next message to publish
                next_message_time = next_record["timestamp"]
                next_message_topic = topic_name

        if next_message_topic is not None:
            self.publish(next_message_topic)
            return True
        elif still_loading:
            return True
        else:
            return False

    def play(self, rate=None, use_s3_directly=None):
        if use_s3_directly is not None:
            self._use_s3_directly = use_s3_directly

        if self._use_s3_directly:
            logger.info("Using S3 directly to fetch data.")
        else:
            logger.info("Using presigned URLs to fetch data.")

        if rate is not None:
            logger.warn("rate is now deprecated.")

        while not rospy.is_shutdown():
            try:
                continue_publishing = self.publish_next_message()
                if not continue_publishing:
                    logger.info("No more messages to publish.")
                    break
            except KeyboardInterrupt:
                self._kill_threads = True
                break



from uuid import UUID

from .common import RESTInterface, output_decorator


class Updater(RESTInterface):
    def __init__(self, config):
        super().__init__(config)

    @output_decorator
    def api_key(self, api_key_id: UUID, data: dict):
        result = self._update_resource(f"apiKeys/{api_key_id}", data)
        return result

    @output_decorator
    def extraction(self, extraction_id: UUID, data: dict):
        result = self._update_resource(f"extractions/{extraction_id}", data)
        return result

    @output_decorator
    def extraction_topic(
        self, extraction_id: UUID, extraction_topic_id: UUID, data: dict
    ):
        result = self._update_resource(
            f"extractions/{extraction_id}/topics/{extraction_topic_id}", data
        )
        return result

    @output_decorator
    def group(self, group_id: UUID, data: dict):
        result = self._update_resource(f"groups/{group_id}", data)
        return result

    @output_decorator
    def group_association(self, group_association_id: UUID, data: dict):
        result = self._update_resource(
            f"groupAssociations/{group_association_id}", data
        )
        return result

    @output_decorator
    def ingestion(self, ingestion_id: UUID, data: dict):
        result = self._update_resource(f"ingestions/{ingestion_id}", data)
        return result

    @output_decorator
    def log(self, log_id, data: dict):
        result = self._update_resource(f"logs/{log_id}", data)
        return result

    @output_decorator
    def message_type(self, message_type_id: UUID, data: dict):
        result = self._update_resource(f"messageTypes/{message_type_id}", data)
        return result

    @output_decorator
    def record(self, timestamp: float, topic_id: UUID, data: dict):
        result = self._update_resource(f"topics/{topic_id}/records/{timestamp}", data)
        return result

    @output_decorator
    def topic(self, topic_id: UUID, data: dict):
        result = self._update_resource(f"topics/{topic_id}", data)
        return result

    @output_decorator
    def user(self, user_id: UUID, data: dict):
        result = self._update_resource(f"users/{user_id}", data)
        return result



import base64
import re
import struct
from collections import OrderedDict


DEBUG = False

# Primitive type handling for ROS builtin types
SIMPLE_TYPES_DICT = {  # see python module struct
    "int8": "b",
    "uint8": "B",
    # Python 2.6 adds in '?' for C99 _Bool, which appears equivalent to an uint8,
    # thus, we use uint8
    "bool": "B",
    "int16": "h",
    "uint16": "H",
    "int32": "i",
    "uint32": "I",
    "int64": "q",
    "uint64": "Q",
    "float32": "f",
    "float64": "d",
    # deprecated
    "char": "B",  # unsigned
    "byte": "b",  # signed
}

# endianess in struct format string
ENDIANNESS = {
    "little": "<",
    "big": ">",
}


class Deserializer:
    def deserialize(self, bytes, offset=0):
        raise NotImplementedError


class CustomROSDeserializer(Deserializer):
    # field regex, field type single word or package/class, optional array and array length in square brackets, single word field name
    # e.g. float64[9] covariance
    # this also ignores any constants defined in the message
    # e.g. float64 PI = 3.141592653589793
    field_regex = re.compile(r"^([\w/]+(\[\d*\])?)\s+(\w+)(\s*#.*)?$")
    array_regex = re.compile(r"^([\w/]+)\[(\d*)\]$")
    trim_cutoff = 1024

    def __init__(self, msg_definitions_dict={}, custom_deserializers={}, trim=False):
        self.msg_definitions_dict = {
            "time": {":is_simple": True, ":size": 8},
            "duration": {":is_simple": True, ":size": 8},
            "string": {
                ":is_simple": False,
                ":size": -1,
            },
            "Header": {
                ":is_simple": False,
                "seq": "uint32",
                "stamp": "time",
                "frame_id": "string",
                ":size": -1,
            },
        }
        self.msg_definitions_dict.update(msg_definitions_dict)
        self.trim = trim
        self.custom_deserializers = {
            "time": self.deserialize_time,
            "duration": self.deserialize_time,
            "string": self.deserialize_string,
            "Header": self.deserialize_header,
        }
        self.custom_deserializers.update(custom_deserializers or {})
        self.byte_order = ENDIANNESS["little"]

    def deserialize(self, message_type, bytes, offset=0):
        res = self.deserialize_message_generic(
            message_type,
            bytes,
            offset=offset,
            byte_order=self.byte_order,
        )[1]
        return res

    # CUSTOM DESERIALIZERS #
    def deserialize_string(
        self, field_type, bytes, offset=0, byte_order=ENDIANNESS["little"]
    ):
        # string is a uint32 length prefix followed by the string
        string_length = struct.unpack(
            f"{byte_order}{SIMPLE_TYPES_DICT['uint32']}", bytes[offset : offset + 4]
        )[0]
        offset += 4

        if self.trim and string_length > self.trim_cutoff:
            return offset + string_length, None

        field_value = bytes[offset : offset + string_length].decode("utf-8")
        return offset + string_length, field_value

    def deserialize_time(
        self, field_type, bytes, offset=0, byte_order=ENDIANNESS["little"]
    ):
        secs, nsecs = struct.unpack(
            f"{byte_order}2{SIMPLE_TYPES_DICT['uint32']}", bytes[offset : offset + 8]
        )
        offset += 8
        return offset, {"secs": secs, "nsecs": nsecs}

    def deserialize_array(
        self,
        field_type,
        bytes,
        package_name=None,
        offset=0,
        byte_order=ENDIANNESS["little"],
    ):
        match = self.array_regex.match(field_type)

        if not match:
            raise ValueError(f"Invalid field type: {field_type}")

        base_type, array_size = match.groups()
        if not array_size:
            # there's a uint32 length prefix to find out the array size
            array_size = struct.unpack(
                f"{byte_order}{SIMPLE_TYPES_DICT['uint32']}", bytes[offset : offset + 4]
            )[0]
            offset += 4
        array_size = int(array_size)

        if base_type in SIMPLE_TYPES_DICT:
            element_size = struct.calcsize(SIMPLE_TYPES_DICT[base_type])
            array_total_size = element_size * array_size

            if self.trim and array_total_size > self.trim_cutoff:
                return offset + array_total_size, None

            # can directly deserialize if it's a simple type
            if base_type == "char":
                field_value = bytes[offset : offset + array_total_size]
                return offset + array_size, base64.b64encode(field_value).decode(
                    "utf-8"
                )

            field_value = struct.unpack(
                f"{byte_order}{array_size}{SIMPLE_TYPES_DICT[base_type]}",
                bytes[offset : offset + array_total_size],
            )
            return offset + array_total_size, list(field_value)

        # otherwise, we need to deserialize each element individually
        field_value = []

        if self.trim and array_size > self.trim_cutoff:
            # save memory/cycles by not deserializing the array elements
            # if we can determine the size of the array elements
            if base_type not in self.msg_definitions_dict:
                base_type = f"{package_name}/{base_type}"
            if self.msg_definitions_dict[base_type][":is_simple"]:
                element_size = self.msg_definitions_dict[base_type][":size"]
                return offset + element_size * array_size, None

        for i in range(array_size):
            offset, element = self.deserialize_message_generic(
                base_type,
                bytes,
                package_name,
                offset,
            )
            field_value.append(element)

        if self.trim and len(field_value) > self.trim_cutoff:
            return offset, None

        return offset, field_value

    def deserialize_header(
        self, field_type, bytes, offset=0, byte_order=ENDIANNESS["little"]
    ):
        seq, stamp_secs, stamp_nsecs, frame_id_len = struct.unpack(
            f"{byte_order}4I", bytes[offset : offset + 16]
        )
        offset += 16
        frame_id = bytes[offset : offset + frame_id_len].decode("utf-8")
        offset += frame_id_len
        return offset, {
            "seq": seq,
            "stamp": {"secs": stamp_secs, "nsecs": stamp_nsecs},
            "frame_id": frame_id,
        }

    def deserialize_message_generic(
        self,
        message_type,
        bytes,
        current_package=None,
        offset=0,
        byte_order=ENDIANNESS["little"],
    ):
        message_type = message_type.split("/", 1)
        if len(message_type) == 1:
            # if message_type is not a fully qualified name, assume it's in the current package
            # there's an exception for the header message, which is in the std_msgs package
            # but we handle it separately
            if message_type[0] in ["string", "Header", "time"]:
                if message_type[0] == "Header":
                    fully_qualified_message_type = "std_msgs/Header"
                else:
                    fully_qualified_message_type = message_type[0]
            else:
                fully_qualified_message_type = current_package + "/" + message_type[0]
            package = current_package
        else:
            package = message_type[0]
            fully_qualified_message_type = package + "/" + message_type[1]

        if fully_qualified_message_type in self.custom_deserializers:
            # avoiding trimming here since we maybe want to allow custom trimming strategies
            return self.custom_deserializers[fully_qualified_message_type](
                field_type=fully_qualified_message_type, bytes=bytes, offset=offset
            )

        if fully_qualified_message_type not in self.msg_definitions_dict:
            raise Exception(f"Message type {fully_qualified_message_type} not found")
        msg_def_dict = self.msg_definitions_dict[fully_qualified_message_type]
        msg_dict = dict()

        for field_name, field_type in msg_def_dict.items():
            if field_name.startswith(":"):
                # internal helper field, skipping
                continue
            if field_type in self.custom_deserializers:
                offset, msg_dict[field_name] = self.custom_deserializers[field_type](
                    field_type=field_type, bytes=bytes, offset=offset
                )
            elif field_type in SIMPLE_TYPES_DICT:
                simple_element_size = struct.calcsize(SIMPLE_TYPES_DICT[field_type])
                msg_dict[field_name] = struct.unpack(
                    f"{byte_order}{SIMPLE_TYPES_DICT[field_type]}",
                    bytes[offset : offset + simple_element_size],
                )[0]
                offset += simple_element_size
            elif self.is_array(field_type):
                offset, msg_dict[field_name] = self.deserialize_array(
                    field_type,
                    bytes,
                    package,
                    offset,
                )
            else:
                offset, msg_dict[field_name] = self.deserialize_message_generic(
                    field_type,
                    bytes,
                    package,
                    offset,
                )
        return offset, msg_dict

    # UTILITY FUNCTIONS #

    def is_array(self, field_type):
        return self.array_regex.match(field_type) is not None

    def get_schema_registry_for_message_type(self, message_type):
        """
        Returns a dictionary of all the schema for message classes needed to serialize/deserialize the given message type.
        """
        if isinstance(message_type, dict):
            breakpoint()
        result = {message_type: list(self.msg_definitions_dict[message_type].items())}
        for field_name, field_type in self.msg_definitions_dict[message_type].items():
            if field_name.startswith(":"):
                # internal helper field, skipping
                continue

            if self.is_array(field_type):
                field_type = self.array_regex.match(field_type).group(1)

            if field_type in result or field_type in SIMPLE_TYPES_DICT:
                continue

            result[field_type] = list(self.msg_definitions_dict[field_type].items())
            dependent_res = self.get_schema_registry_for_message_type(field_type)
            result.update(dependent_res)
        return result

    def _extract_message_definition(self, message_type, message_def):
        """
        format of the further message defined in the message definition
        MSG: <message_name>
        keep parsing until we hit a line that starts with "==="
        """
        msg_name = message_type
        current_line = 0
        lines = message_def.splitlines()
        while current_line < len(lines):
            # fully qualified message name
            msg_name = msg_name
            self.msg_definitions_dict[msg_name] = OrderedDict()
            while current_line < len(lines):
                line = lines[current_line].strip()
                if not line or line.startswith("#"):
                    current_line += 1
                    continue
                if line.startswith("==="):
                    # skip comments
                    current_line += 1
                    while (
                        lines[current_line].strip().startswith("#")
                        or lines[current_line].strip() == ""
                    ):
                        current_line += 1
                    match_obj = re.search(r"^MSG: ([\w/]+)", lines[current_line])
                    if not match_obj:
                        print(
                            "No message name found", lines[current_line], current_line
                        )
                    else:
                        msg_name = match_obj.group(1)
                        current_line += 1
                    break

                match = self.field_regex.match(line)
                if not match:
                    # print(f"Skipping line {current_line}: {line}")
                    current_line += 1
                    continue
                # its still simpler to use split instead of regex groups
                words = line.split()
                field_type = words[0]
                field_name = words[1]

                self.msg_definitions_dict[msg_name][field_name] = field_type
                current_line += 1

    def extract_message_definition(self, message_type, message_def):
        message_delimiter = "================================================================================"
        messages = message_def.split(message_delimiter)
        msg_name_to_msg = OrderedDict()
        for message in messages[::-1][:-1]:
            message_lines = message.splitlines()
            for line in message_lines:
                if line.startswith("MSG:"):
                    msg_name_to_msg[line.split(" ")[1]] = message
        msg_name_to_msg[message_type] = messages[0]
        retry = []
        for msg_type, msg in msg_name_to_msg.items():
            if not self._extract_message_definition_with_size(msg_type, msg):
                retry.append(msg_type)
        for msg_type in retry:
            if not self._extract_message_definition_with_size(
                msg_type, msg_name_to_msg[msg_type]
            ):
                raise Exception(f"Failed to parse message {msg_type}")

    def _extract_message_definition_with_size(self, message_type, message_def):
        msg_name = message_type
        current_line = 0
        lines = message_def.splitlines()
        is_simple = True
        size = 0
        while current_line < len(lines):
            self.msg_definitions_dict[msg_name] = OrderedDict()
            while current_line < len(lines):
                line = lines[current_line].strip()
                if not line or line.startswith("#"):
                    current_line += 1
                    continue

                match = self.field_regex.match(line)
                if not match:
                    current_line += 1
                    continue
                # its still simpler to use split instead of regex groups
                words = line.split()
                field_type = words[0]
                field_name = words[1]
                # check if it's an array of known size
                array_regex = re.compile(r"(.*)\[(\d*)\]")
                match = array_regex.match(field_type)
                if match:
                    if match.groups()[1] != "":
                        # fixed length array
                        base_type = match.groups()[0]
                        if base_type == "string":
                            is_simple = False
                        elif (
                            base_type not in SIMPLE_TYPES_DICT
                            and not self.msg_definitions_dict[base_type][":is_simple"]
                        ):
                            is_simple = False
                        else:
                            if base_type in SIMPLE_TYPES_DICT:
                                size += int(match.groups()[1]) * struct.calcsize(
                                    SIMPLE_TYPES_DICT[base_type]
                                )
                            else:
                                size += (
                                    int(match.groups()[1])
                                    * self.msg_definitions_dict[base_type][":size"]
                                )
                    else:
                        # variable length array
                        is_simple = False
                else:
                    if field_type == "string":
                        is_simple = False
                    if field_type not in SIMPLE_TYPES_DICT:
                        if field_type in self.msg_definitions_dict:
                            if not self.msg_definitions_dict[field_type][":is_simple"]:
                                is_simple = False
                            else:
                                size += self.msg_definitions_dict[field_type][":size"]
                        else:
                            current_package = msg_name.split("/")[0]
                            field_type = current_package + "/" + field_type
                            if field_type not in self.msg_definitions_dict:
                                # clean up
                                del self.msg_definitions_dict[msg_name]
                                return False
                            if not self.msg_definitions_dict[field_type][":is_simple"]:
                                is_simple = False
                            else:
                                size += self.msg_definitions_dict[field_type][":size"]
                    else:
                        # simple types dict
                        size += struct.calcsize(SIMPLE_TYPES_DICT[field_type])

                self.msg_definitions_dict[msg_name][field_name] = field_type
                self.msg_definitions_dict[msg_name][":is_simple"] = is_simple

                self.msg_definitions_dict[msg_name][":size"] = (
                    -1 if not is_simple else size
                )
                current_line += 1
        return True


def make_dict(obj):
    from message_converter import convert_ros_message_to_dictionary

    return convert_ros_message_to_dictionary(obj)


def get_message_class(message_type):
    import importlib
    import sys

    import roslib

    message_class = roslib.message.get_message_class(message_type)
    if message_class is None:
        package, name = message_type.split("/")
        try:
            if f"{package}.msg._{name}" in sys.modules:
                importlib.reload(sys.modules[f"{package}.msg"])
            message_module = importlib.import_module(f"{package}.msg._{name}")
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                f"Failed importing class for message_type {message_type}."
            )
        message_class = getattr(message_module, name)
    return message_class



from collections import OrderedDict
import io
import os
import logging
from uuid import UUID
from enum import Enum
from concurrent.futures import ThreadPoolExecutor

from lqs_client.interface.ark_deserialization import extract_ark_message_from_bytes
from lqs_client.interface.ros_deserialization import CustomROSDeserializer

import numpy as np
from PIL import Image as ImagePIL
from sensor_msgs.msg import Image, CompressedImage

logging.basicConfig(
    level=os.getenv("LQS_LOG_LEVEL") or logging.INFO,
    format="%(asctime)s  (%(levelname)s - %(name)s): %(message)s",
)
logger = logging.getLogger(__name__)


class S3Resource(str, Enum):
    extraction = "extraction"
    ingestion = "ingestion"
    log = "log"
    record = "record"
    topic = "topic"


class Utils:
    def __init__(self, getter, s3, gen, config):
        self._getter = getter
        self._s3 = s3
        self._gen = gen
        self._config = config
        self._cached_ingestions = {}
        self._cached_message_types = {}
        self._cached_topics = {}

        if self._config.get("aws_access_key_id"):
            self._use_s3_directly = True
        else:
            self._use_s3_directly = False

        if self._config.get("verbose"):
            logger.setLevel(logging.DEBUG)

    def upload_part(
        self,
        resource: S3Resource,
        resource_id: UUID,
        file_path: str,
        upload_id: str,
        part_number: int,
        key: str = None,
        part_size: int = 5 * 1024 * 1024,
    ):
        file = open(file_path, "rb")
        file.seek((part_number - 1) * part_size)
        part_data = file.read(part_size)

        logger.debug(f"Uploading part {part_number} ({len(part_data)} bytes)")

        if key is None:
            key = file_path.split("/")[-1]

        r_headers, r_params, r_body = self._s3.upload_part(
            resource=resource,
            resource_id=resource_id,
            key=key,
            part_number=part_number,
            upload_id=upload_id,
            body=part_data,
        )
        e_tag = r_headers["ETag"]
        return {"PartNumber": part_number, "ETag": e_tag}

    def upload(
        self,
        resource: S3Resource,
        resource_id: UUID,
        file_path: str,
        key: str = None,
        part_size: int = 5 * 1024 * 1024,
        max_workers: int = 8,
        exist_ok: bool = False,
    ):
        if key is None:
            key = file_path.split("/")[-1]

        if not exist_ok:
            r_headers, r_params, r_body = self._s3.head_object(
                resource=resource, resource_id=resource_id, key=key
            )
            logger.debug(f"HEAD {resource} ({resource_id}): {key} - {r_body}")
            if r_body["exists"]:
                raise FileExistsError(f"File already exists in S3: {key}")

        logger.debug("Uploading %s to %s", file_path, key)
        logger.debug("Creating multipart upload for %s %s", resource, resource_id)
        (
            _create_mpu_headers,
            _create_mpu_params,
            create_mpu_body,
        ) = self._s3.create_multipart_upload(
            resource=resource, resource_id=resource_id, key=key
        )
        try:
            upload_id = create_mpu_body["InitiateMultipartUploadResult"]["UploadId"]
            logger.debug("Upload ID: %s", upload_id)
        except KeyError as e:
            logger.error("Error creating multipart upload: %s", create_mpu_body)
            raise e

        file = open(file_path, "rb")
        file_size = os.fstat(file.fileno()).st_size
        part_count = (file_size // part_size) + 1
        logger.debug("File size: %s", file_size)
        logger.debug("Part size: %s", part_size)

        parts = []
        futures = []
        logger.debug("Uploading parts with %s workers", max_workers)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            for i in range(part_count):
                part_number = i + 1
                future = executor.submit(
                    self.upload_part,
                    resource=resource,
                    resource_id=resource_id,
                    file_path=file_path,
                    upload_id=upload_id,
                    part_number=part_number,
                    key=key,
                    part_size=part_size,
                )
                futures.append(future)

        for future in futures:
            parts.append(future.result())
            logger.debug("Uploaded part %s of %s", len(parts), part_count)

        logger.debug("Completing multipart upload")
        (
            _complete_mpu_headers,
            _complete_mpu_params,
            complete_mpu_body,
        ) = self._s3.complete_multipart_upload(
            resource=resource,
            resource_id=resource_id,
            key=key,
            upload_id=upload_id,
            parts=parts,
        )
        logger.debug("Completed multipart upload %s", complete_mpu_body)

    # Message Processing

    def get_message_data_from_record(self, record):
        ingestion_id = record["ingestion_id"]
        s3_bucket = record["s3_bucket"]
        s3_key = record["s3_key"]
        if s3_bucket is None or s3_key is None:
            # the record doesn't have the s3 bucket and key, so we need to get them from the ingestion
            if ingestion_id is None:
                raise ValueError(
                    "No S3 bucket or key and no ingestion is available on record."
                )
            elif ingestion_id in self._cached_ingestions:
                ingestion = self._cached_ingestions[ingestion_id]
            elif ingestion_id is not None:
                ingestion = self._getter.ingestion(ingestion_id)["data"]
                self._cached_ingestions[ingestion_id] = ingestion

            s3_bucket = ingestion["s3_bucket"]
            s3_key = ingestion["s3_key"]
            if s3_bucket is None or s3_key is None:
                raise ValueError("No S3 bucket or key available on record's ingestion.")

        message_data = self._s3.get_message_data_from_record(
            record=record, s3_bucket=s3_bucket, s3_key=s3_key, ingestion_id=ingestion_id
        )

        return message_data

    def get_message_class(self, message_type_id):
        if message_type_id in self._cached_message_types:
            message_type = self._cached_message_types[message_type_id]
        else:
            message_type = self._getter.message_type(message_type_id)["data"]
            self._cached_message_types[message_type_id] = message_type

        message_type_name = message_type["name"]
        if "_generated" not in message_type or True:
            self._gen.process_message_definition(
                message_type["definition"], *message_type_name.split("/")
            )
            self._gen.generate_messages()
            self._cached_message_types[message_type_id]["_generated"] = True

        message_class = self._gen.get_message_class(message_type_name)
        return message_class

    def get_ros_message_from_message_data(
        self, message_data, schema_registry, message_type
    ):
        if schema_registry is None:
            return None
        deserializer = CustomROSDeserializer(msg_definitions_dict=schema_registry)
        return deserializer.deserialize(message_type=message_type, bytes=message_data)

    def get_ros_message_from_record(self, record):
        message_data = self.get_message_data_from_record(record=record)
        (
            schema_registry,
            package_name,
            message_class_name,
        ) = self.get_schema_registry_from_record(record=record)
        message_data_dict = self.get_ros_message_from_message_data(
            message_data=message_data,
            schema_registry=schema_registry,
            message_type=f"{package_name}/{message_class_name}",
        )
        if not message_data_dict:
            # get the message class
            message_class = self.get_message_class(
                message_type_id=record["message_type_id"]
            )
            # deserialize the message
            message_data_dict = message_class().deserialize(message_data)
        return message_data_dict, message_class_name

    def get_schema_registry_from_record(self, record):
        topic_id = record["topic_id"]
        if topic_id in self._cached_topics:
            topic = self._cached_topics[topic_id]
        else:
            topic = self._getter.topic(record["topic_id"])["data"]
            self._cached_topics[topic_id] = topic
        if topic.get("type_name") is None:
            package_name, message_class_name = topic["message_type_name"].rsplit("/", 1)
        else:
            package_name, message_class_name = topic["type_name"].rsplit("/", 1)
        schema_registry = topic["type_schema"]
        if schema_registry:
            for k, v in schema_registry.items():
                if isinstance(v, list):
                    # load it as a OrderedDict
                    schema_registry[k] = OrderedDict(v)
        return schema_registry, package_name, message_class_name

    def get_image_from_ros_message(
        self, message, image_mode=None, message_class_name=None
    ):
        # We assume that if the message class name is "Image" or "CompressedImage",
        # then the message type matches the ROS message type.
        if not isinstance(message, dict):
            message_class_name = message.__class__.__name__
            message = {
                "data": message.data,
                "width": message.width,
                "height": message.height,
                "encoding": message.encoding,
            }
            if hasattr(message, "format"):
                message["format"] = message.format

        if isinstance(message["data"], list):
            message["data"] = bytes(message["data"])

        if message_class_name == "Image":
            img_modes = {
                "16UC1": "I;16",
                "mono8": "L",
                "mono16": "I;16",
                "32FC1": "F",
                "8UC1": "L",
                "8UC3": "RGB",
                "rgb8": "RGB",
                "bgr8": "RGB",
                "rgba8": "RGBA",
                "bgra8": "RGBA",
                "bayer_rggb": "L",
                "bayer_rggb8": "L",
                "bayer_gbrg": "L",
                "bayer_gbrg8": "L",
                "bayer_grbg": "L",
                "bayer_grbg8": "L",
                "bayer_bggr": "L",
                "bayer_bggr8": "L",
                "yuv422": "YCbCr",
                "yuv411": "YCbCr",
            }
            if image_mode is None:
                if not message["encoding"]:
                    logger.warn(f"No encoding: {message['encoding']}")
                    return None
                image_mode = img_modes[message["encoding"]]
            img = ImagePIL.frombuffer(
                image_mode,
                (message["width"], message["height"]),
                message["data"],
                "raw",
                image_mode,
                0,
                1,
            )
            if message["encoding"] == "bgr8":
                b, g, r = img.split()
                img = ImagePIL.merge("RGB", (r, g, b))

            if message["encoding"] in ["mono16", "16UC1", "32FC1"]:
                pixels = np.asarray(img)
                pixel_range = np.max(pixels) - np.min(pixels)
                if pixel_range == 0:
                    pixels *= 0
                else:
                    pixels = ((pixels - np.min(pixels)) / pixel_range) * 255.0
                img = ImagePIL.fromarray(pixels)
                img = img.convert("L")
        elif message_class_name == "CompressedImage":
            if message["format"] == "h264":
                import av

                codec = av.CodecContext.create("h264", "r")
                packets = av.packet.Packet(message["data"])
                img = codec.decode(packets)[0].to_image()
            else:
                img = ImagePIL.open(io.BytesIO(message["data"]))
        else:
            raise NotImplementedError(
                f"Message type {type(message)} not supported for image conversion."
            )

        return img

    def process_image(
        self,
        img,
        format="PNG",
        output_path=None,
        resize_width=None,
        resize_height=None,
        resize_scale=None,
    ):
        if (
            resize_width is not None
            or resize_height is not None
            or resize_scale is not None
        ):
            width = img.width
            height = img.height
            if resize_width is not None:
                width = resize_width
            if resize_height is not None:
                height = resize_height
            if resize_scale is not None:
                width = int(width * resize_scale)
                height = int(height * resize_scale)
            img = img.resize((width, height))

        if output_path is not None:
            img.save(output_path, format)
            return None
        else:
            buffered_img = io.BytesIO()
            img.save(buffered_img, format)
            buffered_img.seek(0)
            return buffered_img

    def get_image_from_record(self, record, image_format=None):
        if record.get("format") == "ark":
            return self.get_image_from_ark_record(record, image_format=image_format)
        message, message_class_name = self.get_ros_message_from_record(record=record)
        return self.get_image_from_ros_message(
            message, message_class_name=message_class_name
        )

    def get_processed_image_from_record(self, record):
        img = self.get_image_from_record(record)
        return self.process_image(img)

    def get_image_from_ark_json_msg(self, message, image_format=None):
        data = bytes(message["data"])

        if image_format is None:
            image_format = message["data_format"]

        image_format = image_format.lower()

        if image_format == "depthz16":
            ark_img = np.frombuffer(data, dtype=np.float16)
            pixel_range = np.max(ark_img) - np.min(ark_img)

            if pixel_range != 0:
                ark_img = ((ark_img - np.min(ark_img)) / pixel_range) * 65535.0

            ark_img = ark_img.astype(np.uint16)
            data = ark_img
            image_mode = "I;16"
        elif image_format == "grey":
            image_mode = "L"
        elif image_format == "rgb" or image_format == "bgr":
            image_mode = "RGB"
        elif image_format == "rgba" or image_format == "bgra":
            image_mode = "RGBA"
        elif image_format == "h264":
            import av

            codec = av.CodecContext.create("h264", "r")
            packets = av.packet.Packet(message.data)
            img = codec.decode(packets)[0].to_image()
            return img

        elif image_format.lower() == "nv12":
            width = message["width"]
            height = message["height"]
            data = message["data"]
            Y_size = width * height
            UV_size = (
                Y_size // 2
            )  # There are half as many UV samples as Y samples in NV12

            # Separate Y, U, and V values
            Y = np.array(data[:Y_size], dtype=np.uint8).reshape((height, width))
            UV = np.array(data[Y_size : Y_size + UV_size], dtype=np.uint8).reshape(
                (height // 2, width)
            )

            # Deinterleave U and V components
            U = UV[:, ::2]
            V = UV[:, 1::2]

            # Up-sample U and V components to match Y's dimension
            U_upsampled = np.repeat(np.repeat(U, 2, axis=0), 2, axis=1)
            V_upsampled = np.repeat(np.repeat(V, 2, axis=0), 2, axis=1)

            # Stack the YUV components together to form the YUV image
            YUV = np.stack((Y, U_upsampled, V_upsampled), axis=-1).astype(np.uint8)

            # Convert YUV to RGB
            rgb_image = ImagePIL.fromarray(YUV, "YCbCr").convert("RGB")

            # Save image as png
            return rgb_image
        else:
            raise NotImplementedError(
                f"Message type {image_format} not supported for image conversion."
            )

        img = ImagePIL.frombuffer(
            image_mode,
            (message["width"], message["height"]),
            data,
            "raw",
            image_mode,
            0,
            1,
        )

        if image_format == "bgr":
            b, g, r = img.split()
            img = ImagePIL.merge("RGB", (r, g, b))
        elif image_format == "bgra":
            b, g, r, a = img.split()
            img = ImagePIL.merge("RGBA", (r, g, b, a))
        return img

    def get_ark_message_from_record(self, record):
        message_data = self._s3.get_message_data_for_ark_record(record=record)
        topic_id = record["topic_id"]
        if topic_id in self._cached_topics:
            topic = self._cached_topics[topic_id]
        else:
            topic = self._getter.topic(record["topic_id"])["data"]
            self._cached_topics[topic_id] = topic

        ark_json_msg = extract_ark_message_from_bytes(topic, message_data)
        return ark_json_msg

    def get_image_from_ark_record(self, record, image_format=None):
        ark_json_msg = self.get_ark_message_from_record(record)
        img = self.get_image_from_ark_json_msg(ark_json_msg, image_format=image_format)
        return img



import time
import os
import json
import base64
import pprint
import logging
from enum import Enum

logging.basicConfig(
    level=os.getenv("LQS_LOG_LEVEL") or logging.INFO,
    format="%(asctime)s  (%(levelname)s - %(name)s): %(message)s",
)
logger = logging.getLogger(__name__)

import requests


class ProcessStatus(str, Enum):
    ready = "ready"
    pre_processing = "pre-processing"
    pre_processed = "pre-processed"
    processing = "processing"
    processed = "processed"
    post_processing = "post-processing"
    complete = "complete"


class UserRole(str, Enum):
    viewer = "viewer"
    editor = "editor"
    owner = "owner"


class LogFormat(str, Enum):
    ros = "ros"
    mls = "mls"


def output_decorator(func):
    def wrapper(*args, **kwargs):
        if args[0]._pretty:
            return pprint.pprint(func(*args, **kwargs))
        return func(*args, **kwargs)

    return wrapper


class RESTInterface:
    def __init__(self, config):
        self._api_url = config["api_url"]
        self._api_key_id = config["api_key_id"]
        self._api_key_secret = config["api_key_secret"]
        self._pretty = config["pretty"]
        self._dry_run = config["dry_run"]
        self._retry_count = config["retry_count"]
        self._retry_delay = config["retry_delay"]
        self._retry_aggressive = config["retry_aggressive"]
        self._verify_ssl = config["verify_ssl"]

        auth_header_value = "Bearer " + base64.b64encode(
            bytes(f"{self._api_key_id}:{self._api_key_secret}", "utf-8")
        ).decode("utf-8")

        self._headers = {
            "Authorization": auth_header_value,
            "Content-Type": "application/json",
        }

    def _get_url_param_string(self, args, exclude=[]):
        url_params = ""
        for key, value in args.items():
            if value is not None and key not in ["self"] + exclude:
                if type(value) == dict or type(value) == list:
                    value = json.dumps(value)
                url_params += f"&{key}={value}"
        if len(url_params) > 0:
            url_params = "?" + url_params[1:]
        return url_params

    def _get_payload_data(self, args, exclude=[]):
        payload = {}
        for key, value in args.items():
            if value is not None and key not in ["self"] + exclude:
                payload[key] = value
        return payload

    def _handle_response_data(self, response):
        if response.ok and response.status_code == 204:
            return
        try:
            response_data = response.json()
        except json.decoder.JSONDecodeError:
            raise Exception(f"Error: {response.text}")
        if response.ok:
            return response_data
        else:
            raise Exception(response_data)

    def _handle_retries(self, func, retry_count=None):
        if retry_count is None:
            retry_count = self._retry_count
        for i in range(retry_count + 1):
            try:
                return func()
            except Exception as e:
                if not self._retry_aggressive:
                    lqs_expected_error_codes = [
                        "[BadRequest]",
                        "[Forbidden]",
                        "[NotFound]",
                        "[Conflict]",
                        "[Locked]",
                    ]
                    for code in lqs_expected_error_codes:
                        if code in str(e):
                            raise e
                if retry_count > 0 and i < retry_count:
                    # exponential backoff
                    backoff = self._retry_delay * (2**i)
                    logger.error(f"Error: {e}")
                    logger.info(f"Retrying in {backoff} seconds")
                    time.sleep(backoff)
                else:
                    raise e

    def _get_resource(self, resource_path):
        if self._dry_run:
            logger.info(f"(Dry Run) GET {self._api_url}/{resource_path}")
            return {}

        def make_request():
            r = requests.get(
                f"{self._api_url}/{resource_path}",
                headers=self._headers,
                verify=self._verify_ssl,
            )
            return self._handle_response_data(r)

        return self._handle_retries(make_request)

    def _create_resource(self, resource_path, data):
        if self._dry_run:
            logger.info(f"(Dry Run) POST {self._api_url}/{resource_path}, {data}")
            return {}

        def make_request():
            r = requests.post(
                f"{self._api_url}/{resource_path}",
                data=json.dumps(data),
                headers=self._headers,
                verify=self._verify_ssl,
            )
            return self._handle_response_data(r)

        return self._handle_retries(make_request)

    def _update_resource(self, resource_path, data):
        if self._dry_run:
            logger.info(f"(Dry Run) PATCH {self._api_url}/{resource_path}, {data}")
            return {}

        def make_request():
            r = requests.patch(
                f"{self._api_url}/{resource_path}",
                data=json.dumps(data),
                headers=self._headers,
                verify=self._verify_ssl,
            )
            return self._handle_response_data(r)

        return self._handle_retries(make_request)

    def _delete_resource(self, resource_path):
        if self._dry_run:
            logger.info(f"(Dry Run) DELETE {self._api_url}/{resource_path}")
            return

        def make_request():
            r = requests.delete(
                f"{self._api_url}/{resource_path}",
                headers=self._headers,
                verify=self._verify_ssl,
            )
            return self._handle_response_data(r)

        return self._handle_retries(make_request)



# -*- coding: utf-8 -*-
#
# Software License Agreement (BSD License)
#
# Copyright (c) 2019-2022, Martin Günther (DFKI GmbH) and others
# Copyright (c) 2013-2016, Brandon Alexander
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of this project nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Taken from
# https://github.com/uos/rospy_message_converter/blob/master/src/rospy_message_converter/message_converter.py
# This update hasn't been released, but is needed for critical features.

import logging
import roslib.message
import rospy
import base64
import sys
import copy
import collections

python3 = sys.hexversion > 0x03000000

python_list_types = [list, tuple]

if python3:
    python_string_types = [str, bytes]
    python_int_types = [int]
else:
    python_string_types = [str, unicode]  # noqa
    python_int_types = [int, long]  # noqa

python_float_types = [float]

ros_to_python_type_map = {
    "bool": [bool],
    "float32": copy.deepcopy(python_float_types + python_int_types),
    "float64": copy.deepcopy(python_float_types + python_int_types),
    "int8": copy.deepcopy(python_int_types),
    "int16": copy.deepcopy(python_int_types),
    "int32": copy.deepcopy(python_int_types),
    "int64": copy.deepcopy(python_int_types),
    "uint8": copy.deepcopy(python_int_types),
    "uint16": copy.deepcopy(python_int_types),
    "uint32": copy.deepcopy(python_int_types),
    "uint64": copy.deepcopy(python_int_types),
    "byte": copy.deepcopy(python_int_types),
    "char": copy.deepcopy(python_int_types),
    "string": copy.deepcopy(python_string_types),
}

try:
    import numpy as np

    _ros_to_numpy_type_map = {
        "float32": [np.float32, np.int8, np.int16, np.uint8, np.uint16],
        # don't include int32, because conversion to float may change value:
        # v = np.iinfo(np.int32).max; np.float32(v) != v
        "float64": [
            np.float32,
            np.float64,
            np.int8,
            np.int16,
            np.int32,
            np.uint8,
            np.uint16,
            np.uint32,
        ],
        "int8": [np.int8],
        "int16": [np.int8, np.int16, np.uint8],
        "int32": [np.int8, np.int16, np.int32, np.uint8, np.uint16],
        "int64": [
            np.int8,
            np.int16,
            np.int32,
            np.int64,
            np.uint8,
            np.uint16,
            np.uint32,
        ],
        "uint8": [np.uint8],
        "uint16": [np.uint8, np.uint16],
        "uint32": [np.uint8, np.uint16, np.uint32],
        "uint64": [np.uint8, np.uint16, np.uint32, np.uint64],
        "byte": [np.int8],
        "char": [np.uint8],
    }

    # merge type_maps
    merged = collections.defaultdict(list, ros_to_python_type_map)
    for k, v in _ros_to_numpy_type_map.items():
        merged[k].extend(v)
    ros_to_python_type_map = dict(merged)
except ImportError:
    pass


ros_time_types = ["time", "duration"]
ros_primitive_types = [
    "bool",
    "byte",
    "char",
    "int8",
    "uint8",
    "int16",
    "uint16",
    "int32",
    "uint32",
    "int64",
    "uint64",
    "float32",
    "float64",
    "string",
]
ros_header_types = ["Header", "std_msgs/Header", "roslib/Header"]


def convert_dictionary_to_ros_message(
    message_type,
    dictionary,
    kind="message",
    strict_mode=True,
    check_missing_fields=False,
    check_types=True,
    log_level="error",
):
    """
    Takes in the message type and a Python dictionary and returns a ROS message.
    Example:
        >>> msg_type = "std_msgs/String"
        >>> dict_msg = { "data": "Hello, Robot" }
        >>> convert_dictionary_to_ros_message(msg_type, dict_msg)
        data: "Hello, Robot"
        >>> msg_type = "std_srvs/SetBool"
        >>> dict_msg = { "data": True }
        >>> kind = "request"
        >>> convert_dictionary_to_ros_message(msg_type, dict_msg, kind)
        data: True
    """
    if kind == "message":
        message_class = roslib.message.get_message_class(message_type)
        message = message_class()
    elif kind == "request":
        service_class = roslib.message.get_service_class(message_type)
        message = service_class._request_class()
    elif kind == "response":
        service_class = roslib.message.get_service_class(message_type)
        message = service_class._response_class()
    else:
        raise ValueError('Unknown kind "%s".' % kind)
    message_fields = dict(_get_message_fields(message))

    remaining_message_fields = copy.deepcopy(message_fields)

    if dictionary is None:
        dictionary = {}
    for field_name, field_value in dictionary.items():
        if field_name in message_fields:
            field_type = message_fields[field_name]
            if field_value is not None:
                field_value = _convert_to_ros_type(
                    field_name,
                    field_type,
                    field_value,
                    strict_mode,
                    check_missing_fields,
                    check_types,
                    log_level,
                )
                setattr(message, field_name, field_value)
            del remaining_message_fields[field_name]
        else:
            error_message = 'ROS message type "{0}" has no field named "{1}"'.format(
                message_type, field_name
            )
            if strict_mode:
                raise ValueError(error_message)
            else:
                if log_level not in ["debug", "info", "warning", "error", "critical"]:
                    log_level = "error"
                logger = logging.getLogger("rosout")
                log_func = getattr(logger, log_level)

                log_func("{}! It will be ignored.".format(error_message))

    if check_missing_fields and remaining_message_fields:
        error_message = 'Missing fields "{0}"'.format(remaining_message_fields)
        raise ValueError(error_message)

    return message


def convert_dict_to_message(
    message_class,
    dictionary,
    strict_mode=True,
    check_missing_fields=False,
    check_types=True,
    log_level="error",
):
    message = message_class()
    message_fields = dict(_get_message_fields(message))

    remaining_message_fields = copy.deepcopy(message_fields)

    if dictionary is None:
        dictionary = {}
    for field_name, field_value in dictionary.items():
        if field_name in message_fields:
            field_type = message_fields[field_name]
            if field_value is not None:
                field_value = _convert_to_ros_type(
                    field_name,
                    field_type,
                    field_value,
                    strict_mode,
                    check_missing_fields,
                    check_types,
                    log_level,
                )
                setattr(message, field_name, field_value)
            del remaining_message_fields[field_name]
        else:
            error_message = 'ROS message type "{0}" has no field named "{1}"'.format(
                message_class._type, field_name
            )
            if strict_mode:
                raise ValueError(error_message)
            else:
                if log_level not in ["debug", "info", "warning", "error", "critical"]:
                    log_level = "error"
                logger = logging.getLogger("rosout")
                log_func = getattr(logger, log_level)

                log_func("{}! It will be ignored.".format(error_message))

    if check_missing_fields and remaining_message_fields:
        error_message = 'Missing fields "{0}"'.format(remaining_message_fields)
        raise ValueError(error_message)

    return message


def _convert_to_ros_type(
    field_name,
    field_type,
    field_value,
    strict_mode=True,
    check_missing_fields=False,
    check_types=True,
    log_level="error",
):
    if _is_ros_binary_type(field_type):
        field_value = _convert_to_ros_binary(field_type, field_value)
    elif field_type in ros_time_types:
        field_value = _convert_to_ros_time(field_type, field_value)
    elif field_type in ros_primitive_types:
        # Note: one could also use genpy.message.check_type() here, but:
        # 1. check_type is "not designed to run fast and is meant only for error diagnosis"
        # 2. it doesn't check floats (see ros/genpy#130)
        # 3. it rejects numpy types, although they can be serialized
        if check_types and type(field_value) not in ros_to_python_type_map[field_type]:
            raise TypeError(
                "Field '{0}' has wrong type {1} (valid types: {2})".format(
                    field_name, type(field_value), ros_to_python_type_map[field_type]
                )
            )
        field_value = _convert_to_ros_primitive(field_type, field_value)
    elif _is_field_type_a_primitive_array(field_type):
        field_value = field_value
    elif _is_field_type_an_array(field_type):
        field_value = _convert_to_ros_array(
            field_name,
            field_type,
            field_value,
            strict_mode,
            check_missing_fields,
            check_types,
            log_level,
        )
    else:
        field_value = convert_dictionary_to_ros_message(
            field_type,
            field_value,
            strict_mode=strict_mode,
            check_missing_fields=check_missing_fields,
            check_types=check_types,
            log_level=log_level,
        )
    return field_value


def _convert_to_ros_binary(field_type, field_value):
    if type(field_value) in python_string_types:
        if python3:
            # base64 in python3 added the `validate` arg:
            # If field_value is not properly base64 encoded and there are non-base64-alphabet characters in the input,
            # a binascii.Error will be raised.
            binary_value_as_string = base64.b64decode(field_value, validate=True)
        else:
            # base64 in python2 doesn't have the `validate` arg: characters that are not in the base-64 alphabet are
            # silently discarded, resulting in garbage output.
            binary_value_as_string = base64.b64decode(field_value)
    else:
        binary_value_as_string = bytes(bytearray(field_value))
    return binary_value_as_string


def _convert_to_ros_time(field_type, field_value):
    time = None

    if field_type == "time" and field_value == "now":
        time = rospy.get_rostime()
    else:
        if field_type == "time":
            time = rospy.rostime.Time()
        elif field_type == "duration":
            time = rospy.rostime.Duration()
        if "secs" in field_value and field_value["secs"] is not None:
            setattr(time, "secs", field_value["secs"])
        if "nsecs" in field_value and field_value["nsecs"] is not None:
            setattr(time, "nsecs", field_value["nsecs"])

    return time


def _convert_to_ros_primitive(field_type, field_value):
    # std_msgs/msg/_String.py always calls encode() on python3, so don't do it here
    if field_type == "string" and not python3:
        field_value = field_value.encode("utf-8")
    return field_value


def _convert_to_ros_array(
    field_name,
    field_type,
    list_value,
    strict_mode=True,
    check_missing_fields=False,
    check_types=True,
    log_level="error",
):
    # use index to raise ValueError if '[' not present
    list_type = field_type[: field_type.index("[")]
    return [
        _convert_to_ros_type(
            field_name,
            list_type,
            value,
            strict_mode,
            check_missing_fields,
            check_types,
            log_level,
        )
        for value in list_value
    ]


def convert_ros_message_to_dictionary(message, binary_array_as_bytes=True):
    """
    Takes in a ROS message and returns a Python dictionary.
    Example:
        >>> import std_msgs.msg
        >>> ros_message = std_msgs.msg.UInt32(data=42)
        >>> convert_ros_message_to_dictionary(ros_message)
        {'data': 42}
    """
    dictionary = {}
    message_fields = _get_message_fields(message)
    for field_name, field_type in message_fields:
        field_value = getattr(message, field_name)
        dictionary[field_name] = _convert_from_ros_type(
            field_type, field_value, binary_array_as_bytes
        )

    return dictionary


def _convert_from_ros_type(field_type, field_value, binary_array_as_bytes=True):
    if field_type in ros_primitive_types:
        field_value = _convert_from_ros_primitive(field_type, field_value)
    elif field_type in ros_time_types:
        field_value = _convert_from_ros_time(field_type, field_value)
    elif _is_ros_binary_type(field_type):
        if binary_array_as_bytes:
            field_value = _convert_from_ros_binary(field_type, field_value)
        elif type(field_value) == str:
            field_value = [ord(v) for v in field_value]
        else:
            field_value = list(field_value)
    elif _is_field_type_a_primitive_array(field_type):
        field_value = list(field_value)
    elif _is_field_type_an_array(field_type):
        field_value = _convert_from_ros_array(
            field_type, field_value, binary_array_as_bytes
        )
    else:
        field_value = convert_ros_message_to_dictionary(
            field_value, binary_array_as_bytes
        )

    return field_value


def _is_ros_binary_type(field_type):
    """Checks if the field is a binary array one, fixed size or not
    >>> _is_ros_binary_type("uint8")
    False
    >>> _is_ros_binary_type("uint8[]")
    True
    >>> _is_ros_binary_type("uint8[3]")
    True
    >>> _is_ros_binary_type("char")
    False
    >>> _is_ros_binary_type("char[]")
    True
    >>> _is_ros_binary_type("char[3]")
    True
    """
    return field_type.startswith("uint8[") or field_type.startswith("char[")


def _convert_from_ros_binary(field_type, field_value):
    field_value = base64.b64encode(field_value).decode("utf-8")
    return field_value


def _convert_from_ros_time(field_type, field_value):
    field_value = {"secs": field_value.secs, "nsecs": field_value.nsecs}
    return field_value


def _convert_from_ros_primitive(field_type, field_value):
    # std_msgs/msg/_String.py always calls decode() on python3, so don't do it here
    if field_type == "string" and not python3:
        field_value = field_value.decode("utf-8")
    return field_value


def _convert_from_ros_array(field_type, field_value, binary_array_as_bytes=True):
    # use index to raise ValueError if '[' not present
    list_type = field_type[: field_type.index("[")]
    return [
        _convert_from_ros_type(list_type, value, binary_array_as_bytes)
        for value in field_value
    ]


def _get_message_fields(message):
    return zip(message.__slots__, message._slot_types)


def _is_field_type_an_array(field_type):
    return field_type.find("[") >= 0


def _is_field_type_a_primitive_array(field_type):
    bracket_index = field_type.find("[")
    if bracket_index < 0:
        return False
    else:
        list_type = field_type[:bracket_index]
        return list_type in ros_primitive_types



from collections import OrderedDict
import sys
from .input_stream import InputStream
from typing import Callable, Dict, List, Optional


def read_enum(input_stream: InputStream, enum_name: str, schema_registry: dict):
    values = schema_registry["enums"][enum_name]["values"]
    type = schema_registry["enums"][enum_name]["enum_class"]
    accepted_types = [
        "uint64",
        "uint32",
        "uint16",
        "uint8",
        "int64",
        "int32",
        "int16",
        "int8",
    ]
    assert type in accepted_types, f"enum type must be one of {accepted_types}"
    method_name = f"read_{type}"
    read_fn = getattr(input_stream, method_name)
    value = read_fn()
    return values[value]


def get_read_fn(
    input_stream: InputStream,
    schema_registry: dict,
    field: dict,
    trim,
    trim_size,
):
    type_identifier = field["type"]
    # if its a simple type return the read function
    if type_identifier in [
        "bool",
        "int8",
        "uint8",
        "int16",
        "uint16",
        "int32",
        "uint32",
        "int64",
        "uint64",
        "double",
        "float",
        "string",
        "guid",
    ]:
        return getattr(input_stream, f"read_{type_identifier}")
    if type_identifier == "byte_buffer":
        return lambda: list(input_stream.read_byte_buffer())
    if type_identifier in ["steady_time_point", "system_time_point"]:
        return getattr(input_stream, "read_uint64")
    if type_identifier in ["duration"]:
        return getattr(input_stream, "read_int64")
    if type_identifier in ["enum"]:
        type_name = field["object_type"]
        return lambda: read_enum(
            input_stream, enum_name=type_name, schema_registry=schema_registry
        )
    if type_identifier in ["object"]:
        type_name = field["object_type"]
        read_fn = lambda: deserialize(  # noqa: E731
            input_stream,
            schema_registry,
            type_identifier=type_name,
            trim=trim,
            trim_size=trim_size,
        )

        def wrap():
            # we need to do this since we aren't using read_object which does this logic
            pushed_flags = input_stream.has_more_sections
            input_stream.has_more_sections = False
            res = read_fn()
            input_stream.has_more_sections = pushed_flags
            return res

        return wrap
    if type_identifier in ["arraylist"]:
        # TODO can improve performance for primitive types
        array_size = input_stream.read_array_size(None)
        element_read_func = get_read_fn(
            input_stream,
            schema_registry,
            field["ctr_value_type"],
            trim=trim,
            trim_size=trim_size,
        )
        read_fn = lambda: input_stream.read_array(  # noqa: E731
            read_fn=element_read_func, fixed_array_size=array_size
        )
        if trim and array_size >= trim_size:

            def read_and_ignore():
                read_fn()
                return None

            return read_and_ignore
        return read_fn
    if type_identifier in ["dictionary"]:
        key_type_fn = get_read_fn(
            input_stream,
            schema_registry,
            field["ctr_key_type"],
            trim=trim,
            trim_size=trim_size,
        )
        value_type_fn = get_read_fn(
            input_stream,
            schema_registry,
            field["ctr_value_type"],
            trim=trim,
            trim_size=trim_size,
        )
        dictionary_size = input_stream.read_array_size(None)
        if trim and dictionary_size >= trim_size:

            def read_and_ignore():
                input_stream.read_dictionary(
                    key_type_fn=key_type_fn,
                    value_type_fn=value_type_fn,
                    dictionary_length=dictionary_size,
                )
                return None

            return read_and_ignore
        return lambda: input_stream.read_dictionary(
            key_type_fn=key_type_fn,
            value_type_fn=value_type_fn,
            dictionary_length=dictionary_size,
        )
    if type_identifier in ["array", "variant"]:
        raise NotImplementedError(f"{type_identifier} type not implemented")

    raise NotImplementedError(f"{type_identifier} type not implemented")


def deserialize(
    input_stream: InputStream,
    schema_registry: dict,
    type_identifier: str,
    trim: bool,
    trim_size: int,
    is_enum: Optional[bool] = None,
):
    data_offset = input_stream.buffer.tell()
    if is_enum is None:
        is_enum = type_identifier in schema_registry["enums"]
    if is_enum:
        # TODO check how to deal with bitmask enums
        return read_enum(input_stream, type_identifier, schema_registry)
    else:
        # read the bitsream header
        input_stream.read_bitstream_header()
        res = OrderedDict()
        # read the fields in order
        schema = schema_registry["schemas"][type_identifier]
        for field in schema["fields"]:
            read_fn = get_read_fn(
                input_stream, schema_registry, field, trim=trim, trim_size=trim_size
            )
            field_name = field["name"]
            field_value = read_fn()
            # find size of field
            field_size = sys.getsizeof(field_value)
            if trim and field_size >= trim_size:
                field_value = None
            res[field_name] = field_value

        # import pdb; pdb.set_trace()
        while input_stream.has_more_sections:
            group_id, group_size = input_stream.read_optional_group_header()
            # check if we have a group with this id
            if group_id not in schema["groups"]:
                input_stream.advance(group_size)
            else:
                # process it
                group_start_position = input_stream.buffer.tell()
                schema_group = schema["groups"][group_id]
                for field in schema_group["fields"]:
                    read_fn = get_read_fn(
                        input_stream,
                        schema_registry,
                        field,
                        trim=trim,
                        trim_size=trim_size,
                    )
                    field_name = field["name"]
                    field_value = read_fn()
                    field_size = sys.getsizeof(field_value)
                    if trim and field_size >= trim_size:
                        field_value = None
                    res[field_name] = field_value
                assert (
                    input_stream.buffer.tell() == group_start_position + group_size
                ), f"Warning: group {group_id} started at {group_start_position}, size is {group_size} but the current position is {input_stream.buffer.tell()}"
        data_end_offset = input_stream.buffer.tell()
        return res, data_offset, data_end_offset


def partition_by_identifier(schema_list: List[Dict]):
    return {
        f'{schema["object_namespace"]}::{schema["name"]}': schema
        for schema in schema_list
    }


def get_schema_registry(topic):
    schema_registry = topic["type_schema"]
    schema_registry = {
        k: partition_by_identifier(v) for k, v in schema_registry.items()
    }
    # the keys in enum schemas get stringified because it's over the wire via json
    # we may have to convert it back to the original type
    for enum_name, enum_schema in schema_registry["enums"].items():
        original_key_type = enum_schema["enum_class"]
        if original_key_type == "str":
            continue

        def get_transform_function(original_key_type: str) -> Callable[[str], any]:
            if original_key_type.startswith("int"):
                return int
            raise NotImplementedError()

        transform_function = get_transform_function(original_key_type)

        enum_schema["values"] = {
            transform_function(enum_key): enum_value
            for enum_key, enum_value in enum_schema["values"].items()
        }

    ## similarly we may have to rebuild the groups in the schema
    for schema_name, schema in schema_registry["schemas"].items():
        if "groups" not in schema:
            continue
        schema["groups"] = {v["identifier"]: v for v in schema["groups"].values()}
    return schema_registry


def extract_ark_message_from_bytes(topic, record_bytes):
    schema_registry = get_schema_registry(topic)

    input_stream = InputStream(record_bytes)
    ark_res, data_offset, data_end_offset = deserialize(
        input_stream,
        schema_registry,
        topic["type_name"],
        trim=False,
        trim_size=None,
    )
    return ark_res



from uuid import UUID
from typing import Optional

from .common import RESTInterface, UserRole, LogFormat, output_decorator


class Creator(RESTInterface):
    def __init__(self, config):
        super().__init__(config)

    @output_decorator
    def api_key(self, user_id: UUID, name: str):
        args = locals()
        data = self._get_payload_data(args, [])
        result = self._create_resource(f"apiKeys", data)
        return result

    @output_decorator
    def extraction(
        self, log_id: UUID, name: Optional[str] = None, note: Optional[str] = None
    ):
        args = locals()
        data = self._get_payload_data(args, [])
        result = self._create_resource(f"extractions", data)
        return result

    @output_decorator
    def extraction_presigned_url(self, extraction_id: UUID, method: str, params: dict):
        args = locals()
        data = self._get_payload_data(args, ["extraction_id"])
        result = self._create_resource(
            f"extractions/{extraction_id}/presignedUrls", data
        )
        return result

    @output_decorator
    def extraction_topic(
        self,
        extraction_id: UUID,
        topic_id: UUID,
        start_time: Optional[float] = None,
        end_time: Optional[float] = None,
        frequency: Optional[float] = None,
    ):
        args = locals()
        data = self._get_payload_data(args, ["extraction_id"])
        result = self._create_resource(f"extractions/{extraction_id}/topics", data)
        return result

    @output_decorator
    def group(self, name: str):
        args = locals()
        data = self._get_payload_data(args, [])
        result = self._create_resource(f"groups", data)
        return result

    @output_decorator
    def group_association(self, user_id: UUID, group_id: UUID, role: UserRole):
        args = locals()
        data = self._get_payload_data(args, [])
        result = self._create_resource(f"groupAssociations", data)
        return result

    @output_decorator
    def ingestion(
        self,
        log_id: UUID,
        format: LogFormat,
        s3_bucket: Optional[str] = None,
        s3_key: Optional[str] = None,
        name: Optional[str] = None,
        start_offset: Optional[int] = None,
        end_offset: Optional[int] = None,
        note: Optional[str] = None,
        queued: Optional[bool] = False,
    ):
        args = locals()
        data = self._get_payload_data(args, [])
        result = self._create_resource(f"ingestions", data)
        return result

    @output_decorator
    def ingestion_presigned_url(self, ingestion_id: UUID, method: str, params: dict):
        args = locals()
        data = self._get_payload_data(args, ["ingestion_id"])
        result = self._create_resource(f"ingestions/{ingestion_id}/presignedUrls", data)
        return result

    @output_decorator
    def log(self, group_id: UUID, name: str, note: Optional[str] = None):
        args = locals()
        data = self._get_payload_data(args, [])
        result = self._create_resource(f"logs", data)
        return result

    @output_decorator
    def log_presigned_url(self, log_id: UUID, method: str, params: dict):
        args = locals()
        data = self._get_payload_data(args, ["log_id"])
        result = self._create_resource(f"logs/{log_id}/presignedUrls", data)
        return result

    @output_decorator
    def message_type(
        self,
        group_id: UUID,
        name: str,
        definition: str,
        md5: Optional[str] = None,
        note: Optional[str] = None,
    ):
        args = locals()
        data = self._get_payload_data(args, [])
        result = self._create_resource(f"messageTypes", data)
        return result

    @output_decorator
    def record(
        self,
        timestamp: float,
        topic_id: UUID,
        offset: Optional[int] = None,
        length: Optional[int] = None,
        data_offset: Optional[int] = None,
        data_length: Optional[int] = None,
        errored: Optional[bool] = False,
        archived: Optional[bool] = False,
        error: Optional[dict] = None,
        note: Optional[str] = None,
        chunk_compression: Optional[str] = None,
        chunk_offset: Optional[int] = None,
        chunk_length: Optional[int] = None,
        s3_bucket: Optional[str] = None,
        s3_key: Optional[str] = None,
        format: Optional[LogFormat] = None,
        message_data: Optional[dict] = None,
        message_type_id: Optional[UUID] = None,
        ingestion_id: Optional[UUID] = None,
        payload_bytes: Optional[str] = None,
        payload_dict: Optional[dict] = None,
    ):
        args = locals()
        data = self._get_payload_data(args, ["topic_id"])
        result = self._create_resource(f"topics/{topic_id}/records", data)
        return result

    @output_decorator
    def record_presigned_url(
        self, timestamp: float, topic_id: UUID, method: str, params: dict
    ):
        args = locals()
        data = self._get_payload_data(args, ["timestamp", "topic_id"])
        result = self._create_resource(
            f"topics/{topic_id}/records/{timestamp}/presignedUrls", data
        )
        return result

    @output_decorator
    def topic(
        self,
        name: str,
        log_id: UUID,
        message_type_id: UUID = None,
        ingestion_id: UUID = None,
        type_name: Optional[str] = None,
        type_encoding: Optional[str] = None,
        type_data: Optional[dict] = None,
    ):
        args = locals()
        data = self._get_payload_data(args, [])
        result = self._create_resource(f"topics", data)
        return result

    @output_decorator
    def topic_presigned_url(self, topic_id: UUID, method: str, params: dict):
        args = locals()
        data = self._get_payload_data(args, ["topic_id"])
        result = self._create_resource(f"topics/{topic_id}/presignedUrls", data)
        return result

    @output_decorator
    def user(self, username: str, is_admin: Optional[bool] = False):
        args = locals()
        data = self._get_payload_data(args, [])
        result = self._create_resource(f"users", data)
        return result



import os
import sys
from pathlib import Path
import importlib
import logging

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ros_import_error = False

import rospkg
import rosmsg
import roslib.message
import genpy.generator


LOAD_ENV_MSG_DEFS = False


class DefinitionNotFoundError(Exception):
    pass


class MessageGenerator:
    """Methods for generating message definition classes."""

    def __init__(self, getter, lister, base_path="/tmp"):
        self.getter = getter
        self.lister = lister
        self.message_path = base_path + "/message_generation"

    def list_message_types(self):
        message_types = []
        count = None
        offset = 0

        limit = 100
        max_iters = 100
        current_iter = 0
        max_restarts = 10
        current_restart = 0
        while count is None or count > len(message_types):
            list_message_types_response = self.lister.message_type(
                limit=limit, offset=offset
            )
            new_count = list_message_types_response["count"]
            if count is None:
                count = new_count
            elif count != new_count:
                logger.warn("Message type count changed during load; restarting...")
                message_types = []
                count = None
                offset = 0
                current_iter = 0
                current_restart += 1
                if current_restart > max_restarts:
                    raise Exception("Exceeded max restarts when loading message_types.")
                continue
            message_types += list_message_types_response["data"]
            current_iter += 1
            if current_iter > max_iters:
                raise Exception(
                    f"Exceeded max iterations when loading {count} message_types."
                )
            offset += limit

        logger.debug(f"Loaded {len(message_types)} message types.")
        return message_types

    def setup_message_definitions(self, definition_keys=None):
        download_path = os.path.join(self.message_path, "message_descriptions")
        logger.debug(f"download_path: {download_path}")
        if not os.path.exists(download_path):
            os.makedirs(download_path)

        logger.debug("Listing message types...")
        message_types = self.list_message_types()
        processed_message_types = []
        for message_type in message_types:
            message_type_name = message_type["name"]
            message_type_md5 = message_type["md5"]
            if (definition_keys is not None) and (message_type_name in definition_keys):
                if definition_keys[message_type_name] != message_type_md5:
                    logger.debug(
                        f"Skipping {message_type_name} ({message_type_md5}) based on definition_keys."
                    )
                    continue
            if message_type_name in processed_message_types:
                logger.debug(
                    f"Skipping {message_type_name} ({message_type_md5}) as a type of this name has already been processed."
                )
                continue
            package, name = message_type_name.split("/")
            definition = message_type["definition"]
            self.process_message_definition(definition, package, name)
            processed_message_types.append(message_type_name)

        logger.debug("Generating message classes.")
        self.generate_messages()

    def process_message_definition(self, definition, package, name):
        div = "================================================================================\n"
        defs = definition.split(div)
        main_def = defs[0]
        self.write_message_definition(main_def, package, name)
        for aux_profile in defs[1:]:
            msg_line, aux_def = aux_profile.split("\n", 1)
            assert msg_line[:5] == "MSG: "
            msg_type = msg_line[5:]
            aux_package, aux_name = msg_type.split("/")
            self.write_message_definition(aux_def, aux_package, aux_name)

    def write_message_definition(self, definition, package, name):
        download_path = os.path.join(self.message_path, "message_descriptions")
        target = os.path.join(download_path, f"{package}/msg/")
        if not os.path.exists(os.path.dirname(target)):
            os.makedirs(os.path.dirname(target))
        with open(f"{target}/{name}.msg", "w") as f:
            f.write(definition)

    def generate_package_messages(self, package, package_files, search_path):
        class_dir = os.path.join(self.message_path, f"message_classes/{package}/msg")
        os.makedirs(class_dir, exist_ok=True)

        init_file_path = os.path.join(class_dir, "__init__.py")
        init_file = Path(init_file_path)
        if not init_file.is_file():
            init_file.touch()
        logger.debug(f"package_files: {package_files}")
        with open(init_file_path, "a+") as f:
            for package_file in package_files:
                logger.debug(f"Adding package_file: {package_file}")
                class_name = package_file.split("/")[-1].split(".")[0]
                f.write(f"from ._{class_name} import *\n")

        generator = genpy.generator.MsgGenerator()
        logger.debug(f"Generating {package} messages...")
        result = generator.generate_messages(
            package, package_files, class_dir, search_path
        )
        assert result == 0
        "Error generating messages."

    def get_message_class(self, message_type):
        # TODO: verify that it's ok to remove this
        # message_class = roslib.message.get_message_class(message_type)
        message_class = None
        if message_class is None:
            package, name = message_type.split("/")
            try:
                if f"{package}.msg._{name}" in sys.modules:
                    importlib.reload(sys.modules[f"{package}.msg"])
                    importlib.reload(sys.modules[f"{package}.msg._{name}"])
                message_module = importlib.import_module(f"{package}.msg._{name}")
            except ModuleNotFoundError:
                raise DefinitionNotFoundError(
                    f"Failed importing class for message_type {message_type}."
                )
            message_class = getattr(message_module, name)
        return message_class

    def generate_messages(self):
        logger.debug(f"Generate messages in {self.message_path}.")
        os.makedirs(os.path.join(self.message_path, "message_classes"), exist_ok=True)
        init_file_path = os.path.join(self.message_path, "message_classes/__init__.py")
        init_file = Path(init_file_path)
        if not init_file.is_file():
            init_file.touch()

        logger.debug("Setting up search path...")
        desc_dir = os.path.join(self.message_path, "message_descriptions")
        r = rospkg.RosPack()
        it = rosmsg.iterate_packages(r, ".msg")
        search_path = {}

        if LOAD_ENV_MSG_DEFS:
            logger.debug("Loading environment message definitions...")
            for package, directory in it:
                search_path[package] = [directory]

            layer_path = "/opt/python"
            if os.path.isdir(layer_path):
                for package in os.listdir(layer_path):
                    if package[-5:] == "_msgs":
                        search_path[package] = [f"{layer_path}/{package}/msg"]

        for package in os.listdir(desc_dir):
            search_path[package] = [f"{desc_dir}/{package}/msg"]

        logger.debug(f"Search path: {search_path}")

        logger.debug("Generating local messages...")
        local_path = os.path.join(os.path.dirname(__file__), "..")
        logger.debug(f"local_path: {local_path}")
        local_message_packages = [
            "std_msgs",
            "rosgraph_msgs",
            "actionlib_msgs",
            "diagnostic_msgs",
            "geometry_msgs",
            "nav_msgs",
            "sensor_msgs",
            "shape_msgs",
            "stereo_msgs",
            "trajectory_msgs",
            "visualization_msgs",
        ]
        for package in local_message_packages:
            msg_dir = f"{local_path}/definitions/{package}/msg"
            search_path[package] = [msg_dir]
            package_files = []
            for file in os.listdir(msg_dir):
                if file.endswith(".msg"):
                    package_file = f"{msg_dir}/{file}"
                    package_files.append(package_file)
            logger.debug(f"package: {package}")
            self.generate_package_messages(package, package_files, search_path)

        logger.debug("Generating packages messages...")
        for package in os.listdir(desc_dir):
            msg_dir = f"{desc_dir}/{package}/msg"
            search_path[package] = [msg_dir]
            package_files = []
            for file in os.listdir(msg_dir):
                if file.endswith(".msg"):
                    package_file = f"{msg_dir}/{file}"
                    package_files.append(package_file)
            logger.debug(f"package: {package}")
            self.generate_package_messages(package, package_files, search_path)

        class_dir = os.path.join(self.message_path, "message_classes")
        sys.path.insert(1, class_dir)
        logger.debug(f"PATH: {sys.path}")



from .getter import Getter
from .lister import Lister
from .creator import Creator
from .updater import Updater
from .deleter import Deleter
from .s3 import S3
from .utils import Utils
from .node import Node
from .terminal import Terminal



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lqs-client",
    version="0.1.10",
    author="Nathan Margaglio",
    author_email="nmargaglio@carnegierobotics.com",
    description="LogQS Client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/carnegierobotics/LogQS-Client",
    project_urls={
        "Bug Tracker": "https://github.com/carnegierobotics/LogQS-Client/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    packages=[
        "lqs_client",
        "lqs_client.interface",
        "lqs_client.gen",
        "lqs_client.definitions",
    ],
    python_requires=">=3.6",
    install_requires=[
        "fire==0.4.*",
        "python-dotenv==0.*",
        "requests==2.*",
        "xmltodict==0.13.*",
        "py3rosmsgs==1.18.*",
        "rospkg==1.4.*",
        "tqdm==4.*",
        "boto3==1.26.*",
        "Pillow==9.0.*",
        "numpy==1.22.*",
        "av==9.2.*",
        "lz4==4.0.*",
        "zstd==1.5.*",
        "s3path==0.4.*",
    ],
)
